from tableau_api_lib.api_requests import BaseRequest


class UpdateScheduleRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests updating schedules.
    :param class ts_connection: the Tableau Server connection object
    :param str schedule_name: the new name to give to the schedule
    :param str schedule_priority: an integer value (entered here as a string) between 1 and 100 that determines the
    default priority of the schedule if multiple tasks are pending in the queue; higher numbers have higher priority
    :param str schedule_state: the state for the schedule. Can be either 'Active' or 'Suspended'
    :param str schedule_type: this value (Extract or Subscription) indicates whether the schedule type is an extract or
    a subscription schedule
    :param str schedule_execution_order: parallel to allow jobs associated with this schedule to run at the same time,
    or Serial to require the jobs to run one after the other
    :param str schedule_frequency: the granularity of the schedule. Valid values are:
    [Hourly, Daily, Weekly, or Monthly]
    :param str start_time: the time of day on which scheduled jobs should run (or if the frequency is hourly, on which
    they should start being run), in the format HH:MM:SS (for example, 18:30:00); this value is required for all
    schedule frequencies
    :param str end_time: if the schedule frequency is Hourly, the time of day on which scheduled jobs should stop being
    run, in the format HH:MM:SS (for example, 23:30:00); hourly jobs will occur at the specified intervals between the
    start time and the end time; for other schedule frequencies, this value is not required and if the attribute is
    included, it is ignored
    :param list interval_expression_list: this list of dicts specifies the time interval(s) between jobs on the given
    schedule; the value required here depends on the 'schedule_frequency' value;
    if 'schedule_frequency' = 'Hourly', the interval expression should be either hours="interval" (where "interval" is
    a number [1, 2, 4, 6, 8, 12] in quotes);
    if 'schedule_frequency' = 'Daily', no interval needs to be specified;
    if 'schedule_frequency' = 'Weekly, the interval is weekDay="weekday", where weekday is one of
    ['Sunday', 'Monday', 'Tuesday', etc.] wrapped in quotes;
    if 'schedule_frequency' = 'Monthly', the interval expression is monthDay="day", where day is either the day of the
    month (1-31), or 'LastDay'; in both cases the value is wrapped in quotes.
    """

    def __init__(self,
                 ts_connection,
                 schedule_name=None,
                 schedule_priority=None,
                 schedule_type=None,
                 schedule_state=None,
                 schedule_execution_order=None,
                 schedule_frequency=None,
                 start_time=None,
                 end_time=None,
                 interval_expression_list=None):

        super().__init__(ts_connection)
        self._schedule_name = schedule_name
        self._schedule_priority = schedule_priority
        self._schedule_type = schedule_type
        self._schedule_state = schedule_state
        self._schedule_execution_order = schedule_execution_order
        self._schedule_frequency = schedule_frequency
        self._start_time = start_time
        self._end_time = end_time
        self._interval_expression_list = interval_expression_list
        self._interval_expression_keys = None
        self._interval_expression_values = None
        self._validate_inputs()
        self._validate_state()
        self.base_update_schedule_request()

    @property
    def optional_schedule_param_keys(self):
        return [
            'name',
            'priority',
            'type',
            'state',
            'frequency',
            'executionOrder'
        ]

    @property
    def optional_frequency_param_keys(self):
        return [
            'start',
            'end'
        ]

    @property
    def valid_interval_keys(self):
        return [
            'minutes',
            'hours',
            'weekDay',
            'monthDay'
        ]

    @property
    def valid_schedule_states(self):
        return [
            'active',
            'suspended',
            None
        ]

    def _validate_inputs(self):
        if self._interval_expression_list:
            self._set_interval_expressions()
        else:
            pass

    def _validate_state(self):
        if not self._schedule_state:
            pass
        elif str(self._schedule_state).lower() in self.valid_schedule_states:
            self._schedule_state = self._schedule_state.capitalize()
        else:
            self._invalid_parameter_exception()

    @property
    def optional_schedule_param_values(self):
        return [
            self._schedule_name,
            self._schedule_priority,
            self._schedule_type,
            self._schedule_state,
            self._schedule_frequency,
            self._schedule_execution_order
        ]

    @property
    def optional_frequency_param_values(self):
        return [
            self._start_time,
            self._end_time
        ]

    def _unpack_interval_expressions_list(self, interval_list):
        interval_keys = []
        interval_values = []
        for interval_dict in interval_list:
            for key in interval_dict.keys():
                if key in self.valid_interval_keys:
                    interval_keys.append(key)
                    interval_values.append(interval_dict[key])
                else:
                    self._invalid_parameter_exception()
        return interval_keys, interval_values

    def _set_interval_expressions(self):
        if self._interval_expression_list:
            if any(self._interval_expression_list):
                self._interval_expression_keys, self._interval_expression_values = \
                    self._unpack_interval_expressions_list(
                        self._interval_expression_list)

    @staticmethod
    def _get_parameters_list(param_keys, param_values):
        params_list = []
        for i, key in enumerate(param_keys):
            if param_values[i]:
                params_list.append({key: param_values[i]})
        return params_list

    def base_update_schedule_request(self):
        self._request_body.update({'schedule': {}})
        return self._request_body

    def modified_update_schedule_request(self):
        if any(self.optional_schedule_param_values):
            self._request_body['schedule'].update(
                self._get_parameters_dict(self.optional_schedule_param_keys,
                                          self.optional_schedule_param_values))

        if any(self.optional_frequency_param_values):
            self._request_body['schedule'].update({'frequencyDetails': {}})
            self._request_body['schedule']['frequencyDetails'].update(
                self._get_parameters_dict(self.optional_frequency_param_keys,
                                          self.optional_frequency_param_values))

        if self._interval_expression_list:
            self._request_body['schedule']['frequencyDetails'].update({'intervals': {}})
            self._request_body['schedule']['frequencyDetails']['intervals'].update({
                    'interval': self._get_parameters_list(self._interval_expression_keys,
                                                          self._interval_expression_values)
                })
        return self._request_body

    def get_request(self):
        return self.modified_update_schedule_request()



from tableau_api_lib.api_endpoints import BaseEndpoint


class EncryptionEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 encrypt_extracts=False,
                 decrypt_extracts=False,
                 reencrypt_extracts=False):
        """
        Builds the API endpoint for setting at-rest encryption settings for the active site.
        :param class ts_connection: the Tableau Server connection object
        :param bool encrypt_extracts: True if encrypting all extracts on the active site, False by default
        :param bool decrypt_extracts: True if decrypting all extracts on the active site, False by default
        :param bool reencrypt_extracts: True if reencrypting all extracts with new keys, False by default
        """

        super().__init__(ts_connection)
        self._encrypt_extracts = encrypt_extracts
        self._decrypt_extracts = decrypt_extracts
        self._reencrypt_extracts = reencrypt_extracts
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._encrypt_extracts,
            self._decrypt_extracts,
            self._reencrypt_extracts
        ]

    @property
    def encryption_actions(self):
        return [
            'encrypt-extracts',
            'decrypt-extracts',
            'reencrypt-extracts'
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_encryption_url(self):
        return "{0}/api/{1}/sites/{2}".format(self._connection.server,
                                              self._connection.api_version,
                                              self._connection.site_id)

    @property
    def base_encryption_action_url(self):
        encryption_action = None
        if any(self.mutually_exclusive_params):
            action_index = [index for index, param in enumerate(self.mutually_exclusive_params) if param].pop()
            encryption_action = self.encryption_actions[action_index]
        else:
            self._invalid_parameter_exception()
        return "{0}/{1}".format(self.base_encryption_url,
                                encryption_action)

    def get_endpoint(self):
        return self.base_encryption_action_url



from .base_request import BaseRequest
from .add_datasource_permissions_request import AddDatasourcePermissionsRequest
from .add_datasource_to_favorites_request import AddDatasourceToFavoritesRequest
from .add_datasource_to_schedule_request import AddDatasourceToScheduleRequest
from .add_default_permissions_request import AddDefaultPermissionsRequest
from .add_dqwarning_request import AddDQWarningRequest
from .add_flow_permissions_request import AddFlowPermissionsRequest
from .add_flow_to_schedule_request import AddFlowToScheduleRequest
from .add_project_permissions_request import AddProjectPermissionsRequest
from .add_project_to_favorites_request import AddProjectToFavoritesRequest
from .add_tags_request import AddTagsRequest
from .add_user_to_alert_request import AddUserToAlertRequest
from .add_user_to_group_request import AddUserToGroupRequest
from .add_user_to_site_request import AddUserToSiteRequest
from .add_view_permissions_request import AddViewPermissionsRequest
from .add_view_to_favorites_request import AddViewToFavoritesRequest
from .add_workbook_permissions_request import AddWorkbookPermissionsRequest
from .add_workbook_to_favorites_request import AddWorkbookToFavoritesRequest
from .add_workbook_to_schedule_request import AddWorkbookToScheduleRequest
from .create_extracts_for_workbook_request import CreateExtractsForWorkbookRequest
from .create_group_request import CreateGroupRequest
from .create_project_request import CreateProjectRequest
from .create_schedule_request import CreateScheduleRequest
from .create_site_request import CreateSiteRequest
from .create_subscription_request import CreateSubscriptionRequest
from .empty_request import EmptyRequest
from .graphql_request import GraphqlRequest
from .publish_datasource_request import PublishDatasourceRequest
from .publish_flow_request import PublishFlowRequest
from .publish_workbook_request import PublishWorkbookRequest
from .sign_in_request import SignInRequest
from .switch_site_request import SwitchSiteRequest
from .update_column_request import UpdateColumnRequest
from .update_data_alert_request import UpdateDataAlertRequest
from .update_database_request import UpdateDatabaseRequest
from .update_datasource_connection_request import UpdateDatasourceConnectionRequest
from .update_datasource_request import UpdateDatasourceRequest
from .update_dqwarning_request import UpdateDQWarningRequest
from .update_flow_connection_request import UpdateFlowConnectionRequest
from .update_flow_request import UpdateFlowRequest
from .update_group_request import UpdateGroupRequest
from .update_project_request import UpdateProjectRequest
from .update_schedule_request import UpdateScheduleRequest
from .update_site_request import UpdateSiteRequest
from .update_subscription_request import UpdateSubscriptionRequest
from .update_table_request import UpdateTableRequest
from .update_user_request import UpdateUserRequest
from .update_workbook_connection_request import UpdateWorkbookConnectionRequest
from .update_workbook_now_request import UpdateWorkbookNowRequest
from .update_workbook_request import UpdateWorkbookRequest



class InvalidParameterException(Exception):
    """Raised when an invalid set of parameters are passed to an Endpoint or Request class"""
    def __init__(self, class_name, parameters):
        error_message = """
        "\n{} received an invalid combination of parameters.
         Evaluate the parameters below and correct accordingly:\n{}
         """.format(class_name, parameters)
        super().__init__(error_message)


class InvalidFileTypeException(Exception):
    """Raised when attempting to call publish_workbook() with an invalid file type (valid types are twbx or twb)"""
    def __init__(self, class_name, file_variety, file_extension):
        error_message = ''
        if file_variety.lower() == 'workbook':
            error_message = """
            An error occurred while calling {0}.
            The provided file extension '{2}' is not a valid Tableau {1} file extension.
            Tableau {2} file extensions must be 'twbx' or 'twb'.
            """.format(class_name, file_variety, file_extension)
        elif file_variety.lower() == 'datasource':
            error_message = """
            An error occurred while calling {0}
            The provided file extension '{2}' is not a valid Tableau {1} file extension.
            Tableau {2} file extensions must be 'tds' or 'tdsx'.
            """.format(class_name, file_variety, file_extension)
        super().__init__(error_message)



import os
import json
from urllib3.fields import RequestField
from urllib3.filepost import encode_multipart_formdata

from tableau_api_lib.api_requests import BaseRequest


CHUNK_SIZE = 1024 * 1024 * 5  # 5MB
FILE_SIZE_LIMIT = 1024 * 1024 * 60  # 60MB


class PublishFlowRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests publishing flows.
    :param class ts_connection: the Tableau Server connection object
    :param string flow_name: the name for the flow being published
    :param string flow_file_path: the file path for the flow being published
    :param string project_id: the project ID for the project the flow belongs to
    :param string flow_description: (Optional) s description for the flow being published
    :param list server_address: (Optional) list of string values for each flow connection server address
    :param list port_number: (Optional) list of string values for each flow connection server port number
    :param list connection_username: (Optional) list of string values for each flow connection username
    :param list connection_password: (Optional) list of string values for each flow connection password
    :param list embed_credentials_flag: (Optional) list of boolean values; True if embedding credentials, False if not
    :param list oauth_flag: (Optional) list of boolean values; True if using OAuth, False if not
    """
    def __init__(self,
                 ts_connection,
                 flow_name,
                 flow_file_path,
                 project_id,
                 flow_description=None,
                 server_address=None,
                 port_number=None,
                 connection_username=None,
                 connection_password=None,
                 embed_credentials_flag=False,
                 oauth_flag=False
                 ):
        super().__init__(ts_connection)
        self._flow_name = flow_name
        self._flow_file_path = flow_file_path
        self._project_id = project_id
        self._flow_description = flow_description
        self._server_address = server_address
        self._port_number = port_number
        self._connection_username = connection_username
        self._connection_password = connection_password
        self._embed_credentials_flag = embed_credentials_flag
        self._oauth_flag = oauth_flag
        self._listify_inputs()
        self._file_is_chunked = self._file_requires_chunking()
        self.base_publish_flow_request()

    def _listify_inputs(self):
        if any(self.optional_credentials_param_values) or any(self.optional_connection_param_values):
            if isinstance(self._server_address, str):
                self._server_address = [self._server_address]
            if isinstance(self._port_number, str):
                self._port_number = [self._port_number]
            if isinstance(self._connection_username, str):
                self._connection_username = [self._connection_username]
            if isinstance(self._connection_password, str):
                self._connection_password = [self._connection_password]
            if isinstance(self._embed_credentials_flag, bool) or isinstance(self._embed_credentials_flag, str):
                self._embed_credentials_flag = [self._embed_credentials_flag]
            if isinstance(self._oauth_flag, bool) or isinstance(self._oauth_flag, str):
                self._oauth_flag = [self._oauth_flag]

    @property
    def valid_file_extensions(self):
        return [
            'tfl',
            'tflx'
        ]

    @property
    def optional_flow_param_keys(self):
        return [
            'description'
        ]

    @property
    def optional_connection_param_keys(self):
        return [
            'serverAddress',
            'serverPort'
        ]

    @property
    def optional_credentials_param_keys(self):
        return [
            'name',
            'password',
            'embed',
            'oAuth'
        ]

    @property
    def optional_flow_param_values(self):
        return [
            self._flow_description
        ]

    @property
    def optional_connection_param_values(self):
        return [
            self._server_address,
            self._port_number
        ]

    @property
    def optional_credentials_param_values(self):
        return [
            self._connection_username,
            self._connection_password,
            self._embed_credentials_flag,
            self._oauth_flag
        ]

    def base_publish_flow_request(self):
        self._request_body.update({
            'flow': {
                'name': self._flow_name,
                'project': {'id': self._project_id}
            }
        })
        return self._request_body

    def modified_publish_flow_request(self):
        self._request_body['flow'].update(self._get_parameters_dict(self.optional_flow_param_keys,
                                                                    self.optional_flow_param_values))

        if any(self.optional_connection_param_values) or any(self.optional_credentials_param_values):
            self._request_body['flow'].update({'connections': {'connection': []}})
            for i, _ in enumerate(self._connection_username or [None]):
                self._request_body['flow']['connections']['connection'].append({
                    'serverAddress': self._server_address[i] if self._server_address else None,
                    'portNumber': self._port_number[i] if self._port_number else None,
                    'connectionCredentials': {
                        'name': self._connection_username[i] if self._connection_username else None,
                        'password': self._connection_password[i] if self._connection_password else None,
                        'embed': self._embed_credentials_flag[i] if self._embed_credentials_flag else False
                    }
                })
            if any(self._oauth_flag):
                for i, _ in enumerate(self._oauth_flag):
                    self._request_body['workbook']['connections']['connection'][i]['connectionCredentials'].update({
                        'oAuth': self._oauth_flag[i]
                    })

        return self._request_body

    def _file_requires_chunking(self):
        file_size = os.path.getsize(self._flow_file_path)
        if file_size > FILE_SIZE_LIMIT:
            return True

    def get_flow(self):
        flow_file = os.path.basename(self._flow_file_path)
        with open(self._flow_file_path, 'rb') as f:
            flow_bytes = f.read()
        file_extension = flow_file.split('.')[-1]
        if file_extension in self.valid_file_extensions:
            pass
        else:
            raise Exception('Invalid flow type provided. flow must be a .tfl or .tflx file.')
        return flow_file, flow_bytes

    # testing for chunk upload
    def publish_prep(self, publish_content_type, parameter_dict):
        parameter_dict = parameter_dict if parameter_dict else {'overwrite': 'overwrite=true'}
        filename = os.path.basename(self._flow_file_path)
        file_extension = filename.split('.')[-1]

        if self._file_is_chunked:
            upload_session_id = self._connection.initiate_file_upload().json()['fileUpload']['uploadSessionId']
            parameter_dict.update({'param': 'uploadSessionId={}'.format(upload_session_id)})
            chunks = self.read_chunks(self._flow_file_path)
            for chunk in chunks:
                request, append_content_type = self.chunk_req(chunk)
                file_upload = self._connection.append_to_file_upload(upload_session_id=upload_session_id,
                                                                     payload=request,
                                                                     content_type=append_content_type)

        publishing_headers = self._connection.default_headers.copy()
        publishing_headers.update({'content-type': publish_content_type})
        parameter_dict.update({'flowType': 'flowType={}'.format(file_extension)})
        return publishing_headers, parameter_dict

    @staticmethod
    def read_chunks(file_path):
        with open(file_path, 'rb') as f:
            while True:
                chunked_content = f.read(CHUNK_SIZE)
                if not chunked_content:
                    break
                yield chunked_content

    def chunk_req(self, chunk):
        parts = {'request_payload': (None, '', 'application/json'),
                 'tableau_file': ('file', chunk, 'application/octet-stream')}
        return self._add_multipart(parts)

    @staticmethod
    def _add_multipart(parts):
        mime_multipart_parts = list()
        for name, (filename, data, content_type) in parts.items():
            multipart_part = RequestField(name=name, data=data, filename=filename)
            multipart_part.make_multipart(content_type=content_type)
            mime_multipart_parts.append(multipart_part)
        request, content_type = encode_multipart_formdata(mime_multipart_parts)
        content_type = ''.join(('multipart/mixed',) + content_type.partition(';')[1:])
        return request, content_type

    def _publish_chunked_file_request(self):
        request = self.modified_publish_flow_request()
        parts = {'request_payload': (None, json.dumps(request), 'application/json')}
        return self._add_multipart(parts)

    def _publish_single_file_request(self):
        request = self.modified_publish_flow_request()
        flow_file, flow_bytes = self.get_flow()
        parts = {'request_payload': (None, json.dumps(request), 'application/json'),
                 'tableau_flow': (flow_file, flow_bytes, 'application/octet-stream')}
        return self._add_multipart(parts)

    def get_request(self):
        if self._file_is_chunked:
            return self._publish_chunked_file_request()
        else:
            return self._publish_single_file_request()



"""
Functions to add:
get_all_site_fields
get_sites_dataframe
"""


import pandas as pd
from tableau_api_lib.utils import extract_pages


def get_all_site_fields(conn):
    all_sites = extract_pages(conn.query_sites, parameter_dict={'fields': 'fields=_all_'})
    return all_sites


def get_sites_dataframe(conn, site_names=None, content_urls=None):
    sites_df = pd.DataFrame(get_all_site_fields(conn))
    if site_names:
        sites_df = sites_df[sites_df['name'].isin(site_names)]
    if content_urls:
        sites_df = sites_df[sites_df['contentUrl'].isin(content_urls)]
    return sites_df


def get_active_site_name(conn):
    return conn.query_site().json()['site']['name']


def get_active_site_id(conn):
    return conn.query_site().json()['site']['id']


def get_active_site_content_url(conn):
    return conn.query_site().json()['site']['contentUrl']



from tableau_api_lib.api_endpoints import BaseEndpoint


class WorkbookEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 query_workbooks=False,
                 query_workbook=False,
                 publish_workbook=False,
                 update_workbook=False,
                 delete_workbook=False,
                 update_workbook_connection=False,
                 workbook_id=None,
                 view_id=None,
                 connection_id=None,
                 add_tags=False,
                 delete_tag=False,
                 tag_name=None,
                 create_extracts=False,
                 delete_extracts=False,
                 encryption_flag=False,
                 revision_number=None,
                 downgrade_target_version=None,
                 query_views=False,
                 query_connections=False,
                 query_workbook_preview_img=False,
                 query_workbook_view_preview_img=False,
                 get_workbook_revisions=False,
                 remove_workbook_revision=False,
                 download_workbook=False,
                 download_workbook_pdf=False,
                 download_workbook_revision=False,
                 refresh_workbook=False,
                 get_workbook_downgrade_info=False,
                 parameter_dict=None):
        """
        Builds API endpoints for REST API workbook methods.
        :param class ts_connection: the Tableau Server connection object
        :param bool query_workbooks: True if querying all workbooks; False otherwise
        :param bool query_workbook: True if querying a specific workbook, False otherwise
        :param bool update_workbook_connection: True if updating a specific workbook connection, False otherwise
        :param bool publish_workbook: True if publishing a specific workbook, False otherwise
        :param bool update_workbook: True if updating a specific workbook, False otherwise
        :param bool delete_workbook: True if deleting a specific workbook, False otherwise
        :param str workbook_id: the workbook ID
        :param str view_id: the view ID
        :param str connection_id: the workbook connection ID
        :param bool add_tags: True if adding tags, False otherwise
        :param bool delete_tag: True if deleting a specific tag, False otherwise
        :param str tag_name: the name of the tag
        :param bool create_extracts: True if creating extracts for the workbook, False otherwise
        :param bool delete_extracts: True if deleting extracts for the workbook, False otherwise
        :param bool encryption_flag: True if encrypting extracts for the workbook, False otherwise
        :param str revision_number: the revision number of the workbook revision to download
        :param str downgrade_target_version: the desired downgrade target version for the workbook
        :param bool query_views: True if querying all views, False otherwise
        :param bool query_connections: True if querying all connections, False otherwise
        :param bool query_workbook_preview_img: True if querying a specific preview image, False otherwise
        :param bool query_workbook_view_preview_img: True if querying a specific preview image, False otherwise
        :param bool get_workbook_revisions: True if getting all workbook revisions, False otherwise
        :param bool remove_workbook_revision: True if removing a workbook revision, False otherwise
        :param bool download_workbook: True if downloading workbook content, False otherwise
        :param bool download_workbook_pdf: True if downloading a specific workbook's PDF, False otherwise
        :param bool download_workbook_revision: True if downloading a specific workbook revision, False otherwise
        :param bool refresh_workbook: True if refreshing a specific workbook, False otherwise
        :param bool get_workbook_downgrade_info: True if getting downgrade info for a specific workbook, False otherwise
        :param dict parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the
        literal text that will be appended to the URL endpoint
        """

        super().__init__(ts_connection)
        self._query_workbooks = query_workbooks
        self._query_workbook = query_workbook
        self._publish_workbook = publish_workbook
        self._update_workbook = update_workbook
        self._delete_workbook = delete_workbook
        self._update_workbook_connection = update_workbook_connection
        self._workbook_id = workbook_id
        self._view_id = view_id
        self._connection_id = connection_id
        self._add_tags = add_tags
        self._delete_tag = delete_tag
        self._tag_name = tag_name
        self._create_extracts = create_extracts
        self._delete_extracts = delete_extracts
        self._encryption_flag = encryption_flag
        self._revision_number = revision_number
        self._downgrade_target_version = downgrade_target_version
        self._query_views = query_views
        self._query_connections = query_connections
        self._query_workbook_preview_img = query_workbook_preview_img
        self._query_workbook_view_preview_img = query_workbook_view_preview_img
        self._get_workbook_revisions = get_workbook_revisions
        self._remove_workbook_revision = remove_workbook_revision
        self._download_workbook = download_workbook
        self._download_workbook_pdf = download_workbook_pdf
        self._download_workbook_revision = download_workbook_revision
        self._refresh_workbook = refresh_workbook
        self._get_workbook_downgrade_info = get_workbook_downgrade_info
        self._parameter_dict = parameter_dict
        self._modify_parameter_dict()
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._query_workbooks,
            self._query_workbook,
            self._publish_workbook,
            self._update_workbook,
            self._delete_workbook,
            self._update_workbook_connection,
            self._add_tags,
            self._delete_tag,
            self._create_extracts,
            self._delete_extracts,
            self._query_views,
            self._query_connections,
            self._query_workbook_preview_img,
            self._query_workbook_view_preview_img,
            self._get_workbook_revisions,
            self._remove_workbook_revision,
            self._download_workbook,
            self._download_workbook_pdf,
            self._download_workbook_revision,
            self._refresh_workbook,
            self._get_workbook_downgrade_info
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) > 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    def _modify_parameter_dict(self):
        if self._encryption_flag and not self._parameter_dict:
            self._parameter_dict = {}
        if self._encryption_flag:
            self._parameter_dict.update({"encryption_flag": f"encrypt={self._encryption_flag}"})

    @property
    def base_workbook_url(self):
        return "{0}/api/{1}/sites/{2}/workbooks".format(self._connection.server,
                                                        self._connection.api_version,
                                                        self._connection.site_id)

    @property
    def base_workbook_id_url(self):
        return "{0}/{1}".format(self.base_workbook_url,
                                self._workbook_id)

    @property
    def base_workbook_tags_url(self):
        return "{0}/tags".format(self.base_workbook_id_url)

    @property
    def base_delete_workbook_tag_url(self):
        return "{0}/{1}".format(self.base_workbook_tags_url,
                                self._tag_name)

    @property
    def base_create_extract_url(self):
        return "{0}/createExtract".format(self.base_workbook_id_url)

    @property
    def base_delete_extract_url(self):
        return "{0}/deleteExtract".format(self.base_workbook_id_url)

    @property
    def base_workbook_views_url(self):
        return "{0}/views".format(self.base_workbook_id_url)

    @property
    def base_workbook_connections_url(self):
        return "{0}/connections".format(self.base_workbook_id_url)

    @property
    def base_workbook_connection_id_url(self):
        return "{0}/{1}".format(self.base_workbook_connections_url,
                                self._connection_id)

    @property
    def base_workbook_preview_url(self):
        return "{0}/previewImage".format(self.base_workbook_id_url)

    @property
    def base_workbook_view_preview_url(self):
        return "{0}/{1}/previewImage".format(self.base_workbook_views_url,
                                             self._view_id)

    @property
    def base_workbook_revisions_url(self):
        return "{0}/revisions".format(self.base_workbook_id_url)

    @property
    def base_workbook_downgrade_version_url(self):
        return "{0}/downGradeInfo?productVersion={1}".format(self.base_workbook_id_url,
                                                             self._downgrade_target_version)

    @property
    def base_workbook_revision_removal_url(self):
        return "{0}/{1}".format(self.base_workbook_revisions_url,
                                self._revision_number)

    @property
    def base_workbook_content_url(self):
        return "{0}/content".format(self.base_workbook_id_url)

    @property
    def base_workbook_revision_download_url(self):
        return "{0}/{1}/content".format(self.base_workbook_revisions_url,
                                        self._revision_number)

    @property
    def base_workbook_download_pdf_url(self):
        return "{0}/pdf".format(self.base_workbook_id_url)

    @property
    def base_workbook_refresh_url(self):
        return "{0}/refresh".format(self.base_workbook_id_url)

    @property
    def valid_page_orientations(self):
        return [
            'Portrait',
            'Landscape'
        ]

    @property
    def valid_page_types(self):
        return [
            'A3',
            'A4',
            'A5',
            'B5',
            'Executive',
            'Folio',
            'Ledger',
            'Legal',
            'Letter',
            'Note',
            'Quarto',
            'Tabloid'
        ]

    def _validate_parameter_dict(self):
        if self._parameter_dict:
            parameter_keys = [key.lower() for key in self._parameter_dict.keys()]
            if 'orientation' in parameter_keys:
                if self._parameter_dict['orientation'] in self.valid_page_orientations:
                    pass
                else:
                    raise self._invalid_parameter_exception()
            if 'type' in parameter_keys:
                if self._parameter_dict['type'] in self.valid_page_types:
                    pass
                else:
                    raise self._invalid_parameter_exception()

    def get_endpoint(self):
        url = None
        if self._workbook_id:
            if self._query_workbook:
                url = self.base_workbook_id_url
            elif self._delete_workbook:
                url = self.base_workbook_id_url
            elif self._update_workbook_connection and self._connection_id:
                url = self.base_workbook_connection_id_url
            elif self._update_workbook and not (self._delete_workbook or self._publish_workbook):
                url = self.base_workbook_id_url
            elif self._add_tags:
                url = self.base_workbook_tags_url
            elif self._delete_tag:
                url = self.base_delete_workbook_tag_url
            elif self._create_extracts:
                url = self.base_create_extract_url
            elif self._delete_extracts:
                url = self.base_delete_extract_url
            elif self._query_views:
                url = self.base_workbook_views_url
            elif self._query_connections:
                url = self.base_workbook_connections_url
            elif self._query_workbook_preview_img:
                url = self.base_workbook_preview_url
            elif self._query_workbook_view_preview_img:
                url = self.base_workbook_view_preview_url
            elif self._get_workbook_revisions:
                url = self.base_workbook_revisions_url
            elif self._remove_workbook_revision:
                url = self.base_workbook_revision_removal_url
            elif self._download_workbook:
                url = self.base_workbook_content_url
            elif self._download_workbook_pdf:
                url = self.base_workbook_download_pdf_url
            elif self._download_workbook_revision and self._revision_number:
                url = self.base_workbook_revision_download_url
            elif self._refresh_workbook:
                url = self.base_workbook_refresh_url
            elif self._get_workbook_downgrade_info:
                url = self.base_workbook_downgrade_version_url
            else:
                self._invalid_parameter_exception()
        else:
            url = self.base_workbook_url

        return self._append_url_parameters(url)



from tableau_api_lib.api_endpoints import BaseEndpoint


class GraphqlEndpoint(BaseEndpoint):
    def __init__(self, ts_connection):
        """
        Builds the API endpoint for issuing GraphQL queries to the Metadata API.
        :param class ts_connection: the Tableau Server connection object
        """

        super().__init__(ts_connection)

    @property
    def base_graphql_url(self):
        return "{0}/api/metadata/graphql".format(self._connection.server,
                                                 self._connection.api_version,
                                                 self._connection.site_id)

    def get_endpoint(self):
        return self.base_graphql_url



from tableau_api_lib.api_endpoints import BaseEndpoint


class AuthEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 sign_in=False,
                 sign_out=False,
                 switch_site=False,
                 get_server_info=False,
                 parameter_dict=None):
        """
        Builds the auth endpoint for Tableau Server REST API auth methods.
        :param class ts_connection: the Tableau Server connection object.
        :param bool sign_in: True if signing in, False otherwise.
        :param bool sign_out: True if signing out, False otherwise.
        :param bool switch_site: True if switching site, False otherwise.
        :param bool get_server_info: True if getting server info, False otherwise.
        :param dict parameter_dict: dictionary of URL parameters to append; the value in each key-value pair is the literal
        text that will be appended to the URL endpoint.
        """
        
        super().__init__(ts_connection)
        self._sign_in = sign_in
        self._sign_out = sign_out
        self._switch_site = switch_site
        self._get_server_info = get_server_info
        self._parameter_dict = parameter_dict
        
    @property
    def base_auth_url(self):
        return "{0}/api/{1}/auth".format(self._connection.server, 
                                         self._connection.api_version)
    
    @property
    def base_sign_in_url(self):
        return "{0}/signin".format(self.base_auth_url)
    
    @property
    def base_sign_out_url(self):
        return "{0}/signout".format(self.base_auth_url)
    
    @property
    def base_switch_site_url(self):
        return "{0}/switchSite".format(self.base_auth_url)
    
    @property
    def base_server_info_url(self):
        return "{0}/api/{1}/serverinfo".format(self._connection.server,
                                               self._connection.api_version)
    
    def get_endpoint(self):
        url = None
        if self._sign_in and not (self._sign_out or self._switch_site):
            url = self.base_sign_in_url
        elif self._sign_out and not (self._sign_in or self._switch_site):
            url = self.base_sign_out_url
        elif self._switch_site and not (self._sign_in or self._sign_out):
            url = self.base_switch_site_url
        elif self._get_server_info:
            url = self.base_server_info_url
        
        if url:
            return self._append_url_parameters(url)
        else:
            self._invalid_parameter_exception()   



from tableau_api_lib.api_requests import BaseRequest


class SwitchSiteRequest(BaseRequest):
    """
    Switch site request for generating API api_requests to Tableau Server.
    :param class ts_connection: the Tableau Server connection object
    :param str site_name: the name, AKA content url, of the site you are switching to
    """
    def __init__(self,
                 ts_connection,
                 site_name):

        super().__init__(ts_connection)
        self._site_name = site_name

    def base_switch_site_request(self):
        self._request_body.update({
            'site': {
                'contentUrl': self._site_name
            }
        })
        return self._request_body

    def get_request(self):
        return self.base_switch_site_request()



"""
The functions below enable you to clone the sites from Server A to Server B.
This is particularly useful when doing site exports / imports.
Functions to import: clone_sites
"""


import pandas as pd
import copy

from tableau_api_lib.utils import extract_pages
from tableau_api_lib.exceptions import ContentOverwriteDisabled


"""
1) query sites on source server
2) get site names on source server
3) query sites on target server
4) get site names on target server
5) compare site names between source and target
-> overwrite_policy=None: throw exception if any site names overlap
-> overwrite_policy='overwrite': delete overlapping sites on target server
6) create sites on target server, copying the details from the source server
7) query the newly created sites on the target server
8) merge the source sites and the target sites by name, keeping only the 'state' from the source
9) update the newly created target sites to have their source site's 'state' (Active or Suspended)
    -> return a list of HTTP responses
"""


def get_source_site_df(conn_source, site_names=None) -> pd.DataFrame:
    """
    Query details for all sites on the source server, or only the sites listed in 'site_names'.
    :param class conn_source: the Tableau Server connection object
    :param list site_names: a list of the desired site names whose details will be queried
    :return: Pandas DataFrame
    """
    site_df = pd.DataFrame(extract_pages(conn_source.query_sites))
    if site_names:
        site_df = site_df[site_df['name'].isin(site_names)]
    site_df.fillna(value='-1', inplace=True)
    return site_df


def get_target_site_df(conn_target, site_names=None) -> pd.DataFrame:
    """
    Query details for all sites on the target server, or only the sites listed in 'site_names'.
    :param class conn_target: the Tableau Server connection object
    :param list site_names: a list of the desired site names whose details will be queried
    :return: Pandas DataFrame
    """
    site_df = pd.DataFrame(extract_pages(conn_target.query_sites))
    if site_names:
        site_df = site_df[site_df['name'].isin(site_names)]
    site_df.fillna(value='-1', inplace=True)
    return site_df


def merge_source_and_target_df(source_site_df, target_site_df):
    target_site_subset_df = target_site_df[['id', 'name']]
    merged_df = source_site_df.merge(target_site_subset_df, on='name', how='inner', suffixes=(None, "_target"))
    return merged_df


def get_site_names(site_df) -> list:
    """
    Get the site name for the active site on the specified Tableau Server connection.
    :param DataFrame site_df: the site details DataFrame for the server
    :return: site names
    """
    site_names = list(site_df['name'])
    site_names = [site_name for site_name in site_names if site_name.lower() != 'default']
    return site_names


def get_overlapping_site_names(source_site_df, target_site_df) -> list:
    source_site_names = set(get_site_names(source_site_df))
    target_site_names = set(get_site_names(target_site_df))
    overlapping_site_names = source_site_names.intersection(target_site_names)
    return list(overlapping_site_names)


def create_sites(source_site_df, conn_target):
    print("creating target sites...")
    responses = []
    for i, site in source_site_df.iterrows():
        responses.append(conn_target.create_site(
            site_name=site['name'],
            content_url=site['contentUrl']
        ))
    print("target sites created")
    return responses


def update_sites(conn_target, site_df):
    print("updating target sites...")
    responses = []
    original_site_content_url = conn_target.site_url
    for i, site in site_df.iterrows():
        conn_target.switch_site(content_url=site['contentUrl'])
        responses.append(conn_target.update_site(
            site_id=site['id_target'],
            site_name=site['name'],
            content_url=site['contentUrl'],
            admin_mode=site['adminMode'],
            disable_subscriptions_flag=site['disableSubscriptions'],
            state=site['state'],
            revision_history_enabled_flag=site['revisionHistoryEnabled'],
            revision_limit=site['revisionLimit'],
            subscribe_others_enabled_flag=site['subscribeOthersEnabled'],
            allow_subscription_attachments_flag=site['allowSubscriptionAttachments'],
            guest_access_enabled_flag=site['guestAccessEnabled'],
            cache_warmup_enabled_flag=site['cacheWarmupEnabled'],
            commenting_enabled_flag=site['commentingEnabled'],
            flows_enabled_flag=site['flowsEnabled'],
            extract_encryption_mode=site['extractEncryptionMode'],
            user_quota=site['userQuota']
            # derivedPermissions
            # catalogingEnabled
        ))
    conn_target.switch_site(content_url=original_site_content_url)
    print("target sites updated")
    return responses


def delete_sites(conn, site_details_df, site_names):
    print("deleting overlapping target sites...")
    responses = []
    sites_to_delete = site_details_df[site_details_df['name'].isin(site_names)]
    for i, site in sites_to_delete.iterrows():
        conn_original_state = copy.copy(conn)
        conn.switch_site(content_url=site['contentUrl'])
        responses.append(conn.delete_site(site_name=site['name']))
        conn = conn_original_state
        conn.sign_in()
    print("overlapping target sites deleted")
    return conn


def validate_inputs(overwrite_policy):
    valid_overwrite_policies = [
        None,
        'overwrite'
    ]
    if overwrite_policy in valid_overwrite_policies:
        pass
    else:
        raise ValueError("Invalid overwrite policy provided: '{}'".format(overwrite_policy))


def clone_sites(conn_source, conn_target, site_names=None, overwrite_policy=None):
    validate_inputs(overwrite_policy)
    source_site_df = get_source_site_df(conn_source=conn_source, site_names=site_names)
    target_site_df = get_target_site_df(conn_target=conn_target, site_names=site_names)
    overlapping_site_names = get_overlapping_site_names(source_site_df=source_site_df, target_site_df=target_site_df)
    if any(overlapping_site_names) and not overwrite_policy:
        raise ContentOverwriteDisabled('site')
    if any(overlapping_site_names) and overwrite_policy == 'overwrite':
        conn_target = delete_sites(conn=conn_target, site_details_df=target_site_df, site_names=overlapping_site_names)
    create_sites(source_site_df=source_site_df, conn_target=conn_target)
    target_site_df = get_target_site_df(conn_target=conn_target, site_names=site_names)
    merged_site_df = merge_source_and_target_df(source_site_df=source_site_df, target_site_df=target_site_df)
    updated_sites = update_sites(conn_target=conn_target, site_df=merged_site_df)
    return updated_sites



from tableau_api_lib.api_endpoints import BaseEndpoint


class SchedulesEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 schedule_id=None,
                 create_schedule=False,
                 query_schedules=False,
                 query_extract_schedules=False,
                 update_schedule=False,
                 delete_schedule=False,
                 add_datasource=False,
                 add_workbook=False,
                 add_flow=False,
                 parameter_dict=None):
        """
        Builds API endpoints for REST API schedules methods.
        :param class ts_connection: the Tableau Server connection object
        :param str schedule_id: the schedule ID.
        :param bool create_schedule: True if creating a schedule, False otherwise
        :param bool query_schedules: True if querying all schedules, False otherwise
        :param bool query_extract_schedules: True if querying all extract schedules, False otherwise
        :param bool update_schedule: True if updating a specific schedule, False otherwise
        :param bool delete_schedule: True if deleting a specific schedule, False otherwise
        :param bool add_datasource: True if adding a datasource, False otherwise
        :param bool add_workbook: True if adding a workbook, False otherwise
        :param bool add_flow: True if adding a flow, False otherwise
        :param dict parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the
        literal text that will be appended to the URL endpoint
        """

        super().__init__(ts_connection)
        self._schedule_id = schedule_id
        self._create_schedule = create_schedule
        self._query_schedules = query_schedules
        self._query_extract_schedules = query_extract_schedules
        self._update_schedule = update_schedule
        self._delete_schedule = delete_schedule
        self._add_datasource = add_datasource
        self._add_workbook = add_workbook
        self._add_flow = add_flow
        self._parameter_dict = parameter_dict

    @property
    def mutually_exclusive_params(self):
        return [
            self._create_schedule,
            self._query_schedules,
            self._query_extract_schedules,
            self._update_schedule,
            self._delete_schedule,
            self._add_datasource,
            self._add_workbook,
            self._add_flow
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_schedule_url(self):
        return "{0}/api/{1}/schedules".format(self._connection.server,
                                              self._connection.api_version)

    @property
    def base_site_schedule_id_url(self):
        return "{0}/api/{1}/sites/{2}/schedules/{3}".format(self._connection.server,
                                                            self._connection.api_version,
                                                            self._connection.site_id,
                                                            self._schedule_id)

    @property
    def base_schedule_id_url(self):
        return "{0}/{1}".format(self.base_schedule_url,
                                self._schedule_id)

    @property
    def base_schedule_datasource_url(self):
        return "{0}/datasources".format(self.base_site_schedule_id_url)

    @property
    def base_schedule_workbook_url(self):
        return "{0}/workbooks".format(self.base_site_schedule_id_url)

    @property
    def base_schedule_flow_url(self):
        return "{0}/flows".format(self.base_site_schedule_id_url)

    @property
    def base_schedule_extracts_url(self):
        return "{0}/extracts".format(self.base_schedule_id_url)

    def get_endpoint(self):
        url = None
        if self._schedule_id:
            if self._add_datasource:
                url = self.base_schedule_datasource_url
            elif self._add_workbook:
                url = self.base_schedule_workbook_url
            elif self._add_flow:
                url = self.base_schedule_flow_url
            elif self._update_schedule or self._delete_schedule:
                url = self.base_schedule_id_url
            elif self._query_extract_schedules:
                url = self.base_schedule_extracts_url
            else:
                self._invalid_parameter_exception()
        else:
            url = self.base_schedule_url

        return self._append_url_parameters(url)



import os
import json
from urllib3.fields import RequestField
from urllib3.filepost import encode_multipart_formdata

from tableau_api_lib.api_requests import BaseRequest
from tableau_api_lib.exceptions import InvalidFileTypeException


CHUNK_SIZE = 1024 * 1024 * 5  # 5MB
FILE_SIZE_LIMIT = 1024 * 1024 * 60  # 60MB


class PublishDatasourceRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests publishing datasources.
    :param class ts_connection: the Tableau Server connection object
    :param str datasource_name: the name for the datasource being published
    :param str datasource_file_path: the file path for the datasource being published
    :param str project_id: the project ID for the project the datasource belongs to
    :param str connection_username: if the datasource requires credentials, this value specifies the connection username
    :param str connection_password: if the datasource requires credentials, this value specifies the connection password
    :param bool embed_credentials_flag: True if credentials are to be embedded, False otherwise
    :param bool oauth_flag: True if the data connection username is an OAuth username, False otherwise
    """
    def __init__(self,
                 ts_connection,
                 datasource_name,
                 datasource_file_path,
                 project_id,
                 connection_username=None,
                 connection_password=None,
                 embed_credentials_flag=False,
                 oauth_flag=False
                 ):
        super().__init__(ts_connection)
        self._datasource_name = datasource_name
        self._datasource_file_path = datasource_file_path
        self._project_id = project_id
        self._connection_username = connection_username
        self._connection_password = connection_password
        self._embed_credentials_flag = embed_credentials_flag
        self._oauth_flag = oauth_flag
        self._file_is_chunked = self._file_requires_chunking()
        self.base_publish_datasource_request()

    @property
    def valid_file_extensions(self):
        return [
            'hyper',
            'tds',
            'tdsx',
            'tde'
        ]

    @property
    def optional_credentials_param_keys(self):
        return [
            'name',
            'password',
            'embed',
            'oAuth'
        ]

    @property
    def optional_credentials_param_values(self):
        return [
            self._connection_username,
            self._connection_password,
            self._embed_credentials_flag,
            self._oauth_flag
            # 'true' if self._embed_credentials_flag is True
            # else 'false' if self._embed_credentials_flag is False
            # else None,
            # 'true' if self._oauth_flag is True
            # else 'false' if self._oauth_flag is False
            # else None
        ]

    def base_publish_datasource_request(self):
        self._request_body.update({
            'datasource': {
                'name': self._datasource_name,
                'project': {'id': self._project_id}
            }
        })
        return self._request_body

    def modified_publish_datasource_request(self):
        if any(self.optional_credentials_param_values):
            self._request_body['datasource'].update({'connectionCredentials': {}})
            self._request_body['datasource']['connectionCredentials'].update(
                self._get_parameters_dict(self.optional_credentials_param_keys,
                                          self.optional_credentials_param_values))
        return self._request_body

    def _file_requires_chunking(self):
        file_size = os.path.getsize(self._datasource_file_path)
        if file_size > FILE_SIZE_LIMIT:
            return True

    def get_datasource(self):
        datasource_file = os.path.basename(self._datasource_file_path)
        with open(self._datasource_file_path, 'rb') as f:
            datasource_bytes = f.read()
        file_extension = datasource_file.split('.')[-1]
        if file_extension in self.valid_file_extensions:
            pass
        else:
            raise InvalidFileTypeException(self.__class__.__name__,
                                           file_variety='datasource',
                                           file_extension=file_extension)
        return datasource_file, datasource_bytes

    # testing for chunk upload
    def publish_prep(self, publish_content_type, parameter_dict):
        parameter_dict = parameter_dict if parameter_dict else {'overwrite': 'overwrite=true'}
        filename = os.path.basename(self._datasource_file_path)
        file_extension = filename.split('.')[-1]

        if self._file_is_chunked:
            upload_session_id = self._connection.initiate_file_upload().json()['fileUpload']['uploadSessionId']
            parameter_dict.update({'param': 'uploadSessionId={}'.format(upload_session_id)})
            chunks = self.read_chunks(self._datasource_file_path)
            for chunk in chunks:
                request, append_content_type = self.chunk_req(chunk)
                file_upload = self._connection.append_to_file_upload(upload_session_id=upload_session_id,
                                                                     payload=request,
                                                                     content_type=append_content_type)

        publishing_headers = self._connection.default_headers.copy()
        publishing_headers.update({'content-type': publish_content_type})
        parameter_dict.update({'datasourceType': 'datasourceType={}'.format(file_extension)})
        return publishing_headers, parameter_dict

    @staticmethod
    def read_chunks(file_path):
        with open(file_path, 'rb') as f:
            while True:
                chunked_content = f.read(CHUNK_SIZE)
                if not chunked_content:
                    break
                yield chunked_content

    def chunk_req(self, chunk):
        parts = {'request_payload': (None, '', 'application/json'),
                 'tableau_file': ('file', chunk, 'application/octet-stream')}
        return self._add_multipart(parts)

    @staticmethod
    def _add_multipart(parts):
        mime_multipart_parts = list()
        for name, (filename, data, content_type) in parts.items():
            multipart_part = RequestField(name=name, data=data, filename=filename)
            multipart_part.make_multipart(content_type=content_type)
            mime_multipart_parts.append(multipart_part)
        request, content_type = encode_multipart_formdata(mime_multipart_parts)
        content_type = ''.join(('multipart/mixed',) + content_type.partition(';')[1:])
        return request, content_type

    def _publish_chunked_file_request(self):
        request = self.modified_publish_datasource_request()
        parts = {'request_payload': (None, json.dumps(request), 'application/json')}
        return self._add_multipart(parts)

    def _publish_single_file_request(self):
        request = self.modified_publish_datasource_request()
        datasource_file, datasource_bytes = self.get_datasource()
        parts = {'request_payload': (None, json.dumps(request), 'application/json'),
                 'tableau_datasource': (datasource_file, datasource_bytes, 'application/octet-stream')}
        return self._add_multipart(parts)

    def get_request(self):
        if self._file_is_chunked:
            return self._publish_chunked_file_request()
        else:
            return self._publish_single_file_request()



import urllib


def flatten_dict_column(df, keys, col_name):
    for key in keys:
        df[col_name + '_' + key] = df[col_name].apply(lambda col: col[key])
    df.drop(columns=[col_name], inplace=True)
    return df


def get_server_netloc(server_address):
    return urllib.parse.urlparse(server_address).netloc



from tableau_api_lib.api_requests import BaseRequest


class UpdateDQWarningRequest(BaseRequest):
    def __init__(self,
                 ts_connection,
                 warning_type,
                 message,
                 status=None):
        """
        Builds the request body for Tableau Server REST API requests updating data quality warnings.
        :param class ts_connection: the Tableau Server connection object
        :param string warning_type: the warning type [Deprecated, Warning, Stale data, Under maintenance]
        :param string message: a custom message accompanying the data warning
        :param boolean status: (optional) toggles the warning on or off [True for on, False for off]
        """

        super().__init__(ts_connection)
        self._warning_type = warning_type
        self._message = message
        self._status = status
        self._request_body = {'dataQualityWarning': {}}
        self._validate_inputs()

    @property
    def valid_warning_types(self):
        return[
            'Deprecated',
            'Warning',
            'Stale data',
            'Under maintenance'
        ]

    def _validate_inputs(self):
        valid = True
        if self._warning_type not in self.valid_warning_types:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def optional_param_keys(self):
        return [
            'type',
            'message',
            'isActive'
        ]

    @property
    def optional_param_values(self):
        return [
            self._warning_type,
            self._message,
            self._status
        ]

    def base_update_column_request(self):
        self._request_body['dataQualityWarning'].update(self._get_parameters_dict(self.optional_param_keys,
                                                                                  self.optional_param_values))
        return self._request_body

    def get_request(self):
        return self.base_update_column_request()



from tableau_api_lib.api_endpoints import BaseEndpoint


class ColumnEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 query_column=False,
                 query_columns=False,
                 update_column=False,
                 remove_column=False,
                 table_id=None,
                 column_id=None):
        """
        Builds the API endpoint for interacting with database table columns.
        :param class ts_connection: the Tableau Server connection object
        :param bool query_column: True if querying a specific column by ID, False by default
        :param bool query_columns: True if querying all columns for the active site, False by default
        :param bool update_column: True if updating a specific column, False by default
        :param bool remove_column: True if removing a specific column, False by default
        :param str table_id: the database table ID
        :param str column_id: the column ID
        """

        super().__init__(ts_connection)
        self._query_column = query_column
        self._query_columns = query_columns
        self._update_column = update_column
        self._remove_column = remove_column
        self._table_id = table_id
        self._column_id = column_id
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._query_column,
            self._query_columns,
            self._update_column,
            self._remove_column
        ]

    def _validate_inputs(self):
        if sum(self.mutually_exclusive_params) == 1:
            pass
        else:
            self._invalid_parameter_exception()

    @property
    def base_column_url(self):
        return "{0}/api/{1}/sites/{2}/tables/{3}/columns".format(self._connection.server,
                                                                 self._connection.api_version,
                                                                 self._connection.site_id,
                                                                 self._table_id)

    @property
    def base_column_id_url(self):
        return "{0}/{1}".format(self.base_column_url,
                                self._column_id)

    def get_endpoint(self):
        if self._column_id and self._table_id:
            url = self.base_column_id_url
        else:
            url = self.base_column_url
        if url:
            return url
        else:
            self._invalid_parameter_exception()



"""
Helper functions for querying REST API data for groups
"""


import pandas as pd
from tableau_api_lib.utils import extract_pages
from tableau_api_lib.utils.querying import get_users_dataframe


def get_all_group_fields(conn):
    all_groups = extract_pages(conn.query_groups, parameter_dict={'fields': 'fields=_all_'})
    return all_groups


def get_group_users(conn, group_id):
    all_group_users = extract_pages(conn.get_users_in_group,
                                    content_id=group_id,
                                    parameter_dict={'fields': 'fields=_all_'})
    return all_group_users


def get_all_group_names(conn):
    all_groups = get_all_group_fields(conn)
    all_groupnames = [group['name'] for group in all_groups]
    return all_groupnames


def get_all_group_domain_names(conn):
    all_groups = get_all_group_fields(conn)
    all_group_roles = [group['domain']['name'] for group in all_groups]
    return all_group_roles


def get_groups_dataframe(conn):
    groups_df = pd.DataFrame(get_all_group_fields(conn))
    groups_df['domain'] = groups_df['domain'].apply(lambda x: x['name'])
    return groups_df


def get_group_users_dataframe(conn, group_id):
    group_users_df = pd.DataFrame(get_group_users(conn, group_id))
    users_df = get_users_dataframe(conn)
    users_columns = list(users_df.columns)
    all_group_users = group_users_df.merge(users_df, how='left', left_on='id', right_on='id', suffixes=('delete', None))
    return all_group_users[users_columns]



from tableau_api_lib.api_requests import BaseRequest


class UpdateProjectRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests updating projects.
    :param class ts_connection: the Tableau Server connection object
    :param str project_name: (optional) the new name for the project
    :param str project_description: (optional) the new description for the project
    :param str content_permissions: (optional) the new permissions setting for the project; Specify LockedToProject to
    lock permissions so that users cannot overwrite the default permissions set for the project, or specify
    ManagedByOwner to allow users to manage permissions for content that they own; For more information, see Lock
    Content Permissions to the Project; The default value is ManagedByOwner
    :param str parent_project_id: (optional) the identifier of the parent project. Use this option to create project
    hierarchies;
    Note: To update a project without changing its placement in the project hierarchy, omit the parentProjectId
    attribute. To move a project to the top of the project hierarchy, provide an empty string ("") for the
    parentProjectId attribute
    """
    def __init__(self,
                 ts_connection,
                 project_name=None,
                 project_description=None,
                 content_permissions=None,
                 parent_project_id=None):
        super().__init__(ts_connection)
        self._project_name = project_name
        self._project_description = project_description
        self._content_permissions = content_permissions
        self._parent_project_id = parent_project_id
        self.base_create_project_request()

    @property
    def optional_project_param_keys(self):
        return [
            'name',
            'parentProjectId',
            'description',
            'contentPermissions'
        ]

    @property
    def optional_project_param_values(self):
        return [
            self._project_name,
            self._parent_project_id,
            self._project_description,
            self._content_permissions
        ]

    def base_create_project_request(self):
        self._request_body.update({'project': {}})
        return self._request_body

    def modified_create_project_request(self):
        self._request_body['project'].update(
            self._get_parameters_dict(
                self.optional_project_param_keys,
                self.optional_project_param_values))
        if self._parent_project_id == "":
            self._request_body['project'].update({'parentProjectId': ''})
        return self._request_body

    def get_request(self):
        return self.modified_create_project_request()



from tableau_api_lib.api_endpoints import BaseEndpoint


class DQWarningEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 warning_id=None,
                 content_type=None,
                 content_id=None,
                 add_warning=False,
                 query_by_id=False,
                 query_by_content=False,
                 update_warning=False,
                 delete_by_id=False,
                 delete_by_content=False):
        """
        Builds the API endpoint for creating and interacting with data quality warnings.
        :param class ts_connection: the Tableau Server connection object
        :param str warning_id: the data quality warning ID
        :param str content_type: the content type receiving the data quality warning
        [datasource, table, flow, or database]
        :param str content_id: the content ID for the specific content receiving the data quality warning
        :param bool add_warning: True if adding a data quality warning, False by default
        :param bool query_by_id: True if querying a data quality warning by ID, False by default
        :param bool query_by_content: True if querying a data quality warning for a piece of content, False by default
        :param bool update_warning: True if updating a data quality warning, False by default
        :param bool delete_by_id: True if deleting a data quality warning by ID, False by default
        :param bool delete_by_content: True if deleting a data quality warning for a piece of content, False by default
        """

        super().__init__(ts_connection)
        self._warning_id = warning_id
        self._content_type = str(content_type).lower() if content_type else None
        self._content_id = content_id
        self._add_warning = add_warning
        self._query_by_id = query_by_id
        self._query_by_content = query_by_content
        self._update_warning = update_warning
        self._delete_by_id = delete_by_id
        self._delete_by_content = delete_by_content
        self._validate_inputs()

    @property
    def valid_content_types(self):
        return [
            'database',
            'table',
            'datasource',
            'flow'
        ]

    @property
    def mutually_exclusive_params(self):
        return [
            self._add_warning,
            self._query_by_id,
            self._query_by_content,
            self._update_warning,
            self._delete_by_id,
            self._delete_by_content
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if self._content_type not in self.valid_content_types:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_warning_url(self):
        return "{0}/api/{1}/sites/{2}/dataQualityWarnings".format(self._connection.server,
                                                                  self._connection.api_version,
                                                                  self._connection.site_id)

    @property
    def base_warning_id_url(self):
        return "{0}/{1}".format(self.base_warning_url,
                                self._warning_id)

    @property
    def base_content_type_url(self):
        return "{0}/contentType/{1}/contentLuid/{2}".format(self.base_warning_url,
                                                            self._content_type,
                                                            self._content_id)

    def get_endpoint(self):
        url = None
        if any([self._add_warning, self._query_by_content, self._delete_by_content]):
            url = self.base_content_type_url
        elif any([self._query_by_id, self._delete_by_id, self._update_warning]):
            url = self.base_warning_id_url
        if url:
            return url
        else:
            self._invalid_parameter_exception()



"""
Functions to add:
get_all_project_fields
get_projects_dataframe
"""


import pandas as pd
from tableau_api_lib.utils import extract_pages


def get_all_project_fields(conn):
    all_projects = extract_pages(conn.query_projects, parameter_dict={'fields': 'fields=_all_'})
    return all_projects


def get_projects_dataframe(conn):
    projects_df = pd.DataFrame(get_all_project_fields(conn))
    return projects_df



from tableau_api_lib.api_endpoints import BaseEndpoint


class SiteEndpoint(BaseEndpoint):
    def __init__(self, 
                 ts_connection, 
                 site_id=None,
                 site_name=None,
                 content_url=None,
                 create_site=False,
                 update_site=False,
                 delete_site=False,
                 query_site=False,
                 query_sites=False,
                 query_views=False,
                 get_users=False,
                 get_groups=False,
                 get_recently_viewed=False,
                 get_recommendations=False,
                 add_user=False,
                 add_group=False,
                 remove_user=False,
                 remove_group=False,
                 user_id=None,
                 group_id=None,
                 parameter_dict=None):
        """
        Builds API endpoints for REST API site methods.
        :param class ts_connection: the Tableau Server connection object
        :param str site_id: the site ID
        :param str site_name: the site name (only required if deleting a site by name)
        :param str content_url: the site name (only required if deleting a site by contentUrl)
        :param bool create_site: True if creating a site, False otherwise
        :param bool update_site: True if updating a specific site, False otherwise
        :param bool delete_site: True if deleting a specific site, False otherwise
        :param bool query_site: True if querying a specific site, False otherwise
        :param bool query_sites: True if querying all sites on the site, False otherwise
        :param bool get_users: True if getting all users, False otherwise
        :param bool get_groups: True if getting all groups, False otherwise
        :param bool get_recently_viewed: True if getting recently viewed content, False otherwise
        :param bool get_recommendations: True if getting recommendations, False otherwise
        :param bool add_user: True if adding a user, False otherwise
        :param bool add_group: True if adding a group, False otherwise
        :param bool remove_user: True if removing a specific user, False otherwise
        :param bool remove_group: True if removing a specific group, False otherwise
        :param str user_id: the user ID
        :param str group_id: the group ID
        :param dict parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the
        literal text that will be appended to the URL endpoint
        """
        
        super().__init__(ts_connection)
        self._site_id = site_id
        self._site_name = site_name
        self._content_url = content_url
        self._create_site = create_site
        self._update_site = update_site
        self._delete_site = delete_site
        self._query_site = query_site
        self._query_sites = query_sites
        self._query_views = query_views
        self._get_users = get_users
        self._get_groups = get_groups
        self._get_recently_viewed = get_recently_viewed
        self._get_recommendations = get_recommendations
        self._add_user = add_user
        self._add_group = add_group
        self._remove_user = remove_user
        self._remove_group = remove_group
        self._user_id = user_id
        self._group_id = group_id
        self._parameter_dict = parameter_dict
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._create_site,
            self._update_site,
            self._delete_site,
            self._query_site,
            self._query_sites,
            self._query_views,
            self._get_users,
            self._get_groups,
            self._get_recently_viewed,
            self._get_recommendations,
            self._add_user,
            self._add_group,
            self._remove_user,
            self._remove_group
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()
        
    @property
    def base_site_url(self):
        return "{0}/api/{1}/sites".format(self._connection.server, 
                                          self._connection.api_version)
    
    @property
    def base_site_id_url(self):
        return "{0}/{1}".format(self.base_site_url, 
                                self._site_id)
    
    @property
    def base_site_views_url(self):
        return "{0}/views".format(self.base_site_id_url)
    
    @property
    def base_site_user_url(self):
        return "{0}/users".format(self.base_site_id_url)

    @property
    def base_site_content_recent_url(self):
        return "{0}/content/recent".format(self.base_site_id_url)

    @property
    def base_site_recommendations_url(self):
        return "{0}/recommendations".format(self.base_site_id_url)
    
    @property
    def base_site_user_id_url(self):
        return "{0}/{1}".format(self.base_site_user_url,
                                self._user_id)
    
    @property
    def base_site_group_url(self):
        return "{0}/groups".format(self.base_site_id_url)
    
    @property
    def base_site_group_id_url(self):
        return "{0}/{1}".format(self.base_site_group_url,
                                self._group_id)

    @property
    def base_delete_site_url(self):
        if self._site_id:
            return "{0}/{1}".format(self.base_site_url,
                                    self._site_id)
        elif self._site_name:
            return "{0}/{1}?key=name".format(self.base_site_url,
                                             self._site_name)
        elif self._content_url:
            return "{0}/{1}?key=contentUrl".format(self.base_site_url,
                                                   self._content_url)
    
    def get_endpoint(self):
        url = None
        if not self._delete_site and (self._site_id or self._user_id or self._group_id):
            if self._update_site:
                url = self.base_site_id_url
            elif self._get_users:
                url = self.base_site_user_url
            elif self._add_user:
                url = self.base_site_user_url
            elif self._get_groups:
                url = self.base_site_group_url
            elif self._get_recently_viewed:
                url = self.base_site_content_recent_url
            elif self._get_recommendations:
                url = self.base_site_recommendations_url
            elif self._add_group:
                url = self.base_site_group_url
            elif self._remove_user and self._user_id and self._site_id:
                url = self.base_site_user_id_url
            elif self._remove_group and self._group_id and self._site_id:
                url = self.base_site_group_id_url
            elif self._query_site:
                url = self.base_site_id_url
            elif self._query_views:
                url = self.base_site_views_url
            else:
                self._invalid_parameter_exception()
        elif self._delete_site and (self._site_id or self._site_name or self._content_url):
            url = self.base_delete_site_url
        else:
            url = self.base_site_url

        return self._append_url_parameters(url)



"""
The functions below enable you to clone the projects from Server A to Server B.
This is particularly useful when doing site exports / imports.
Functions to import: clone_projects
"""


import pandas as pd

from tableau_api_lib.utils import extract_pages
from tableau_api_lib.utils.querying import get_users_dataframe
from tableau_api_lib.exceptions import ContentOverwriteDisabled


def get_source_project_df(conn_source, project_names=None):
    """
    Query details for all projects on the source site, or only the projects listed in 'project_names'.
    :param class conn_source: the Tableau Server connection object
    :param list project_names: a list of the desired project names whose details will be queried
    :return: Pandas DataFrame
    """
    project_df = pd.DataFrame(extract_pages(conn_source.query_projects))
    if project_names:
        project_df = project_df[project_df['name'].isin(project_names)]
    project_df['source_owner_id'] = project_df['owner'].apply(extract_project_owner_id)
    project_df.rename(columns={
        'id': 'source_project_id',
        'name': 'source_project_name',
        'description': 'source_project_description',
        'parentProjectId': 'source_project_parent_id'
    }, inplace=True)
    return project_df


def get_target_project_df(conn_target, project_names=None):
    """
    Query details for all projects on the target site, or only the projects listed in 'project_names'.
    :param class conn_target: the Tableau Server connection object
    :param list project_names: a list of the desired project names whose details will be queried
    :return: Pandas DataFrame
    """
    project_df = pd.DataFrame(extract_pages(conn_target.query_projects))
    if project_names:
        project_df = project_df[project_df['name'].isin(project_names)]
    project_df['target_owner_id'] = project_df['owner'].apply(extract_project_owner_id)
    project_df.rename(columns={
        'id': 'target_project_id',
        'name': 'target_project_name',
        'description': 'target_project_description',
        'parentProjectId': 'target_project_parent_id'
    }, inplace=True)
    return project_df


def get_project_names(project_df, column_name) -> list:
    """
    Get the project names for the active site on the specified Tableau Server connection.
    :param DataFrame project_df: the project details DataFrame for the server
    :param str column_name: the name of the column containing project names [source_project_name, target_project_name]
    :return: project names
    """
    project_names = list(project_df[column_name])
    project_names = [project_name for project_name in project_names]
    return project_names


def get_overlapping_project_names(source_project_df, target_project_df) -> list:
    source_project_names = set(get_project_names(source_project_df, 'source_project_name'))
    target_project_names = set(get_project_names(target_project_df, 'target_project_name'))
    overlapping_project_names = source_project_names.intersection(target_project_names)
    return list(overlapping_project_names)


def extract_project_owner_id(owner_dict, key='id'):
    """
    Extract the user ID for the owner of the project.
    :param dict owner_dict: dictionary containing the project owner ID key-value pair
    :param str key: the key for the project owner ID key-value pair
    :return str: project_owner_id
    """
    project_owner_id = owner_dict[key]
    return project_owner_id


def get_source_user_df(conn):
    """
    Get a DataFrame with user details from the source site.
    :param class conn:
    :return DataFrame: source_users_df
    """
    cols = ['email', 'fullName', 'id']
    source_users_df = get_users_dataframe(conn)
    source_users_df = source_users_df[cols]
    source_users_df.rename(columns={
        'email': 'source_email',
        'fullName': 'source_full_name',
        'id': 'source_owner_id'
    }, inplace=True)
    return source_users_df


def get_target_user_df(conn):
    """
    Get a DataFrame with user details from the target site.
    :param class conn:
    :return DataFrame: target_users_df
    """
    cols = ['email', 'fullName', 'id']
    target_users_df = get_users_dataframe(conn)
    target_users_df = target_users_df[cols]
    target_users_df.rename(columns={
        'email': 'target_email',
        'fullName': 'target_full_name',
        'id': 'target_owner_id'
    }, inplace=True)
    return target_users_df


def join_source_projects_and_users(source_project_df, source_user_df, target_user_df):
    """
    Join the source project details, source user details, and target user details.
    :param DataFrame source_project_df: the source project details
    :param DataFrame source_user_df: the source user details
    :param DataFrame target_user_df: the target user details
    :return DataFrame: joined_df
    """
    joined_df = source_project_df.merge(source_user_df, left_on='source_owner_id', right_on='source_owner_id')
    joined_df = joined_df.merge(target_user_df, left_on='source_email', right_on='target_email', how='left')
    return joined_df


def get_source_to_target_df(conn_source, conn_target, project_names=None):
    """
    Get a DataFrame with source project details, source user details, and source target details combined.
    :param class conn_source: the source site connection
    :param class conn_target: the target site connection
    :param list project_names: (optional) the list of desired projects to clone
    :return DataFrame: project_details_df
    """
    project_df = get_source_project_df(conn_source, project_names)
    source_user_df = get_source_user_df(conn_source)
    target_user_df = get_target_user_df(conn_target)
    project_details_df = join_source_projects_and_users(project_df, source_user_df, target_user_df)
    return project_details_df


def get_child_projects_df(project_details_df):
    """
    Get details for projects that have an associated parent project ID.
    :param DataFrame project_details_df: the details for all projects
    :return DataFrame: child_projects_df
    """
    child_projects_df = project_details_df[project_details_df['source_project_parent_id'].notna()]
    return child_projects_df


def get_parent_projects_df(project_details_df):
    """
    Get details for projects that have an associated child project.
    :param DataFrame project_details_df: the details for all projects
    :return DataFrame: parent_projects_df
    """
    parent_projects_df = project_details_df[project_details_df['source_project_parent_id'].isna()]
    return parent_projects_df


def add_target_project_ids(project_details_df, conn_target):
    """
    Join the project details with the target site's project IDs.
    :param DataFrame project_details_df: the project details DataFrame
    :param class conn_target: the target server's connection
    :return: DataFrame
    """
    target_project_df = get_target_project_df(conn_target)[['target_project_id', 'target_project_name']]
    target_project_df.rename(columns={'id': 'target_project_id'}, inplace=True)
    project_details_df = project_details_df.merge(target_project_df,
                                                  left_on='source_project_name',
                                                  right_on='target_project_name',
                                                  how='left')
    return project_details_df


def add_source_parent_project_names(project_details_df):
    """
    Join the parent project names to the details for each project.
    :param DataFrame project_details_df: the project details DataFrame
    :return: DataFrame
    """
    temp_df = project_details_df[['source_project_name', 'source_project_id']].copy()
    temp_df.rename(columns={'source_project_name': 'source_project_parent_name'}, inplace=True)
    project_details_df = project_details_df.merge(temp_df,
                                                  left_on='source_project_parent_id',
                                                  right_on='source_project_id',
                                                  how='left',
                                                  suffixes=(None, '_delete',))
    return project_details_df.drop(columns=['source_project_id_delete'])


def add_target_parent_project_ids(project_details_df):
    """
    Join the parent project IDs to the details for each project.
    :param DataFrame project_details_df: the project details DataFrame
    :return: DataFrame
    """
    temp_df = project_details_df[['target_project_name', 'target_project_id']].copy()
    temp_df.rename(columns={'target_project_id': 'target_project_parent_id'}, inplace=True)
    project_details_df = project_details_df.merge(temp_df,
                                                  left_on='source_project_parent_name',
                                                  right_on='target_project_name',
                                                  how='left',
                                                  suffixes=(None, '_delete'))
    return project_details_df.drop(columns=['target_project_name_delete'])


def create_final_target_df(conn_source, conn_target):
    """
    Create a project details DataFrame intended for use after creating projects on the target server.
    :param class conn_source: the source server connection
    :param class conn_target: the target server connection
    :return: DataFrame
    """
    project_details_df = get_source_to_target_df(conn_source, conn_target)
    project_details_df = add_target_project_ids(project_details_df, conn_target=conn_target)
    project_details_df = add_source_parent_project_names(project_details_df)
    project_details_df = add_target_parent_project_ids(project_details_df)
    return project_details_df


def create_projects(project_details_df, conn_target):
    """
    Create projects on the target server, based on project details from a source server.
    :param DataFrame project_details_df: the project details DataFrame
    :param class conn_target: the target server connection
    :return: list of HTTP responses
    """
    responses = []
    for i, _ in enumerate(project_details_df['source_project_name']):
        response = conn_target.create_project(
            project_name=project_details_df['source_project_name'][i],
            project_description=project_details_df['source_project_description'][i],
            content_permissions=project_details_df['contentPermissions'][i],
            parent_project_id=None)
        responses.append(response)
    return responses


def update_project_hierarchies(project_details_df, conn_target):
    """
    Update projects on the target server to match the project hierarchies defined on the source server.
    :param DataFrame project_details_df: the project details DataFrame
    :param class conn_target: the target server connection
    :return: list of HTTP responses
    """
    responses = []
    child_projects_df = get_child_projects_df(project_details_df)
    for index, row in child_projects_df.iterrows():
        response = conn_target.update_project(
            project_id=row['target_project_id'],
            parent_project_id=row['target_project_parent_id']
        )
        responses.append(response)
    return responses


def validate_inputs(overwrite_policy):
    valid_overwrite_policies = [
        None,
        'overwrite'
    ]
    if overwrite_policy in valid_overwrite_policies:
        pass
    else:
        raise ValueError("Invalid overwrite policy provided: '{}'".format(overwrite_policy))


def delete_projects(conn, project_details_df, project_names):
    print("deleting overlapping target projects...")
    responses = []
    projects_to_delete = project_details_df[project_details_df['target_project_name'].isin(project_names)]
    for i, project in projects_to_delete.iterrows():
        responses.append(conn.delete_project(project_id=project['target_project_id']))
    print("overlapping target projects deleted")
    return conn


def clone_projects(conn_source, conn_target, project_names=None, overwrite_policy=None):
    """
    Clones projects from the source server to the target server.
    :param class conn_source: the source server connection
    :param class conn_target: the target server connection
    :param list project_names: (optional) a list of project names to clone; specifying no names clones all projects
    :param str overwrite_policy: (optional) set to 'overwrite' to overwrite content; defaults to not overwriting
    :return: None
    """
    validate_inputs(overwrite_policy)
    source_project_df = get_source_project_df(conn_source=conn_source, project_names=project_names)
    target_project_df = get_target_project_df(conn_target=conn_target, project_names=project_names)
    overlapping_project_names = get_overlapping_project_names(source_project_df=source_project_df,
                                                              target_project_df=target_project_df)
    if any(overlapping_project_names) and not overwrite_policy:
        raise ContentOverwriteDisabled('project')
    if any(overlapping_project_names) and overwrite_policy == 'overwrite':
        delete_projects(conn_target, project_details_df=target_project_df, project_names=overlapping_project_names)
    project_details_df = get_source_to_target_df(conn_source, conn_target, project_names)
    create_projects(project_details_df, conn_target)
    update_project_hierarchies(create_final_target_df(conn_source, conn_target), conn_target)



from tableau_api_lib.api_endpoints import BaseEndpoint


class ProjectEndpoint(BaseEndpoint):
    def __init__(self, 
                 ts_connection,
                 create_project=False,
                 query_projects=False,
                 update_project=False,
                 delete_project=False,
                 project_id=None, 
                 parameter_dict=None):
        """
        Builds API endpoints for REST API project methods.
        :param class ts_connection: the Tableau Server connection object
        :param bool create_project: True if creating a project, False otherwise
        :param bool query_projects: True if querying all projects, False otherwise
        :param bool update_project: True if updating a specific project, False otherwise
        :param bool delete_project: True if deleting a specific project, False otherwise
        :param str project_id: the project ID
        :param dict parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the
        literal text that will be appended to the URL endpoint
        """
        
        super().__init__(ts_connection)
        self._create_project = create_project
        self._query_projects = query_projects
        self._update_project = update_project
        self._delete_project = delete_project
        self._project_id = project_id
        self._parameter_dict = parameter_dict
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._create_project,
            self._query_projects,
            self._update_project,
            self._delete_project
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()
        
    @property
    def base_project_url(self):
        return "{0}/api/{1}/sites/{2}/projects".format(self._connection.server, 
                                                       self._connection.api_version, 
                                                       self._connection.site_id)
    
    @property
    def base_project_id_url(self):
        return "{0}/{1}".format(self.base_project_url, 
                                self._project_id)
        
    def get_endpoint(self):
        if self._project_id:
            url = self.base_project_id_url
        else:
            url = self.base_project_url
        if url:
            return self._append_url_parameters(url)
        else:
            self._invalid_parameter_exception()



"""
Functions to add:
get_all_schedule_fields
get_schedules_dataframe
"""


import pandas as pd
from tableau_api_lib.utils import extract_pages


def get_all_schedule_fields(conn):
    all_schedules = extract_pages(conn.query_schedules, parameter_dict={'fields': 'fields=_all_'})
    return all_schedules


def get_schedules_dataframe(conn):
    schedules_df = pd.DataFrame(get_all_schedule_fields(conn))
    return schedules_df



from tableau_api_lib.api_requests import BaseRequest


class AddProjectToFavoritesRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests adding project to favorites.
    :param class ts_connection: the Tableau Server connection object
    :param str favorite_label: the text label to assign to the favorite item
    :param str project_id: the project ID
    """
    def __init__(self,
                 ts_connection,
                 favorite_label,
                 project_id):
        super().__init__(ts_connection)
        self._favorite_label = favorite_label
        self._project_id = project_id

    def base_add_favorites_request(self):
        self._request_body.update({
            'favorite': {
                'label': self._favorite_label,
                'project': {'id': self._project_id}
            }
        })
        return self._request_body

    def get_request(self):
        return self.base_add_favorites_request()



"""
Logical Flow:
1) get users from source connection
2) get users from target connection
3) if a user mapping file is provided, join it to source users
4) get overlapping users
    4a) if user mapping file is provided, do an inner join on target users
5) if there are any overlapping users and no overwrite policy is set, raise an exception
6) if there are any overlapping users and the overwrite policy is 'overwrite', delete the overlapping target users
7) for each user on the source connection, create a new user on the target connection as defined by the mapping file
    7a) if no mapping file is provided, create a new user on the target connection matching the source connection

*Note the limitation of Tableau Online: cannot update user full names... so they are stuck with a busted full name

"""


import pandas as pd

from tableau_api_lib.exceptions import ContentOverwriteDisabled
from tableau_api_lib.utils.querying.users import get_users_dataframe


def verify_mapping_file_columns(mapping_file_columns, required_columns) -> None:
    """
    Throws an error if required columns in the mapping file are not present.
    :param list mapping_file_columns: list of columns present in the mapping file
    :param list required_columns: list of required columns for mapping users
    :return: None
    """
    missing_columns = [col for col in required_columns if col not in mapping_file_columns]
    if any(missing_columns):
        raise ValueError(""""
        The mapping file provided is missing the following required columns: {}
        """.format(missing_columns))
    pass


def get_mapping_file_df(mapping_file_path) -> pd.DataFrame:
    """
    Creates a Pandas DataFrame from the CSV file located in the mapping_file_path.
    :param str mapping_file_path: the file path to the mapping file CSV
    :return: Pandas DataFrame
    """
    required_columns = ['source_site_name', 'source_username', 'target_site_name', 'target_username']
    try:
        mapping_file_df = pd.read_csv(mapping_file_path)
        verify_mapping_file_columns(list(mapping_file_df.columns), required_columns)
        return mapping_file_df
    except FileNotFoundError:
        raise FileNotFoundError("The mapping file '{}' could not be found. Please verify the file path provided.")


def get_source_user_df(conn_source, usernames=None) -> pd.DataFrame:
    """
    Creates a Pandas DataFrame populated with data for users on the source Tableau Server connection.
    :param class conn_source: the source Tableau Server connection
    :param list usernames: (optional) a subset of users; if specified, only these users will appear in the Dataframe
    :return: Pandas DataFrame
    """
    user_df = get_users_dataframe(conn_source)
    if usernames:
        user_df = user_df[user_df['name'].isin(usernames)]
    user_df.rename(columns={
        'name': 'source_username',
        'fullName': 'source_full_name',
        'email': 'source_email',
        'siteRole': 'source_site_role',
        'authSetting': 'source_auth_setting'
    }, inplace=True, errors='ignore')
    if 'source_auth_setting' not in user_df.columns:
        user_df['source_auth_setting'] = 'ServerDefault'
    return user_df


def get_target_user_df(conn_target, usernames=None) -> pd.DataFrame:
    """
    Creates a Pandas DataFrame populated with data for users on the target Tableau Server connection.
    :param class conn_target: the source Tableau Server connection
    :param list usernames: (optional) a subset of users; if specified, only these users will appear in the Dataframe
    :return: Pandas DataFrame
    """
    user_df = get_users_dataframe(conn_target)
    if usernames:
        user_df = user_df[user_df['name'].isin(usernames)]
    user_df.rename(columns={
        'name': 'target_username',
        'fullName': 'target_full_name',
        'email': 'target_email',
        'siteRole': 'target_site_role',
        'authSetting': 'target_auth_setting'
    }, inplace=True, errors='ignore')
    return user_df


def get_overlapping_usernames(source_user_df,
                              target_user_df,
                              mapping_file_path=None) -> list:
    """
    Creates a list of usernames that appear in the source connection and the target connection.
    :param pd.DataFrame source_user_df: a Pandas DataFrame populated with source connection user details
    :param pd.DataFrame target_user_df: a Pandas DataFrame populated with target connection user details
    :param str mapping_file_path: (optional) the file path for the mapping file CSV
    :return: overlapping_usernames
    """
    if mapping_file_path:
        source_usernames = set(source_user_df['target_username'])
    else:
        source_usernames = set(source_user_df['source_username'])
    target_usernames = set(target_user_df['target_username'])
    overlapping_usernames = source_usernames.intersection(target_usernames)
    return list(overlapping_usernames)


def get_mapped_user_df(user_df, mapping_file_path) -> pd.DataFrame:
    """
    Creates a Pandas DataFrame that combines source user details with user mapping details.
    :param pd.DataFrame user_df: a Pandas Dataframe with source connection user details
    :param str mapping_file_path: the file path for the mapping file CSV
    :return: user_df
    """
    if mapping_file_path:
        mapping_file_df = get_mapping_file_df(mapping_file_path)
        user_df = user_df.merge(mapping_file_df,
                                left_on='source_username',
                                right_on='source_username',
                                suffixes=('_delete', None))
    return user_df


def delete_users(conn, target_user_df) -> list:
    """
    Removes users from the specified connection.
    :param class conn:
    :param pd.DataFrame target_user_df:
    :return: list of HTTP responses
    """
    print("removing overlapping target users...")
    if conn.username in list(target_user_df['target_username']):
        target_user_df = target_user_df[~target_user_df['target_username'].isin([conn.username])]
        print("preserving user '{}', which was used to authenticate into the target Tableau Server."
              .format(conn.username))
    user_ids = target_user_df['id']
    responses = [conn.remove_user_from_site(user_id) for user_id in user_ids]
    [print(response) for response in responses]
    print("overlapping target users removed")
    return responses


def create_users(conn_target, source_user_df, mapping_file_path, server_type) -> list:
    """
    Create users on the target server, based on user details from the source server.
    :param class conn_target: the target server connection
    :param DataFrame source_user_df: the source user DataFrame containing details for the users being created
    :param mapping_file_path: the file path to the user mapping file CSV
    :param server_type: (optional) the product variety ['tableau_server', 'tableau_online']
    :return: list of HTTP responses
    """
    responses = []
    field_prefix = get_field_prefix(mapping_file_path)
    source_user_df = source_user_df[~source_user_df[field_prefix + 'site_role'].isin(['ServerAdministrator'])]
    if server_type == 'tableau_server':
        for index, row in source_user_df.iterrows():
            response = conn_target.add_user_to_site(
                user_name=row[field_prefix + 'username'],
                site_role=row[field_prefix + 'site_role'],
                auth_setting=row[field_prefix + 'auth_setting'])
            responses.append(response)
    if server_type == 'tableau_online':
        for index, row in source_user_df.iterrows():
            response = conn_target.add_user_to_site(
                user_name=row[field_prefix + 'email'],
                site_role=row[field_prefix + 'site_role'],
                auth_setting=row[field_prefix + 'auth_setting'])
            responses.append(response)
    for response in responses:
        print(response.content)
    return responses


def reset_connection(conn) -> object:
    """
    Resets the connection to Tableau Server by signing out and signing in again.
    :param class conn: the Tableau Server connection to reset
    :return: TableauServerConnection
    """
    conn.sign_out()
    conn.sign_in()
    return conn


def get_field_prefix(mapping_file_path) -> str:
    """
    Get the field prefix, depending on whether or not a mapping_file_path was provided.
    :param str mapping_file_path: the file path to the user mapping file
    :return: field_prefix
    """
    if mapping_file_path:
        field_prefix = 'target_'
    else:
        field_prefix = 'source_'
    return field_prefix


def update_users(conn_target, source_user_df, mapping_file_path) -> list:
    """
    Updates users on the target connection to mirror values on the source connection.
    :param class conn_target: the target Tableau Server connection
    :param pd.DataFrame source_user_df: a Pandas DataFrame with source user details
    :param str mapping_file_path: the file path to the user mapping file
    :return: list of HTTP responses
    """
    print("updating users...")
    reset_connection(conn_target)
    target_user_df = get_target_user_df(conn_target)
    target_user_df = target_user_df[~target_user_df['target_username'].isin([conn_target.username])]
    field_prefix = get_field_prefix(mapping_file_path)
    combined_user_df = target_user_df.merge(source_user_df,
                                            left_on='target_username',
                                            right_on=field_prefix + 'username',
                                            suffixes=('_update', None))
    responses = []
    for index, row in combined_user_df.iterrows():
        response = conn_target.update_user(
            user_id=row['id_update'],
            new_full_name=row[field_prefix + 'full_name'],
            new_email=row[field_prefix + 'email'],
            new_site_role=row[field_prefix + 'site_role'],
            new_auth_setting=row[field_prefix + 'auth_setting'])
        responses.append(response)
    for response in responses:
        print(response.content)
    print("users updated")
    return responses


def fill_expected_columns(source_user_df):
    """
    Populates any missing 'expected_columns' with None value if they do not yet exist.
    :param pd.DataFrame source_user_df: a Pandas DataFrame with source user details
    :return: source_user_df
    """
    expected_columns = ['source_email', 'source_full_name', 'source_site_role', 'source_auth_setting']
    source_columns = source_user_df.columns
    for expected_column in expected_columns:
        if expected_column not in source_columns:
            source_user_df[expected_column] = None
    return source_user_df


def process_overlapping_usernames(conn_target,
                                  target_user_df,
                                  overlapping_usernames,
                                  overwrite_policy) -> None:
    """
    Processes overlapping usernames depending on the overwrite policy specified.
    :param class conn_target: the target Tableau Server connection
    :param pd.DataFrame target_user_df: a Pandas DataFrame with target user details
    :param list overlapping_usernames: a list of usernames present on both the source and target connections
    :param str overwrite_policy: must be set to 'overwrite' to enable overwriting existing content
    :return:
    """
    if any(overlapping_usernames) and not overwrite_policy:
        raise ContentOverwriteDisabled('users')
    if any(overlapping_usernames) and overwrite_policy == 'overwrite':
        delete_users(conn_target, target_user_df[target_user_df['target_username'].isin(overlapping_usernames)])
    pass


def clone_users(conn_source,
                conn_target,
                server_type='tableau_server',
                usernames=None,
                mapping_file_path=None,
                overwrite_policy=None) -> list:
    """
    Clones users from the source Tableau Server connection to the target Tableau Server connection.
    :param class conn_source: the source Tableau Server connection
    :param class conn_target: the target Tableau Server connection
    :param str server_type: (optional) the product variety ['tableau_server', 'tableau_online']
    :param list usernames: (optional) a list of users to clone; if provided, only these users are cloned
    :param str mapping_file_path: (optional) the path to a file mapping users from source to target
    :param  overwrite_policy: (optional) must be set to 'overwrite' to enable overwriting existing content
    :return:
    """
    source_user_df = get_source_user_df(conn_source, usernames)
    source_user_df = get_mapped_user_df(source_user_df, mapping_file_path)
    source_user_df = fill_expected_columns(source_user_df)
    target_user_df = get_target_user_df(conn_target, usernames)
    overlapping_usernames = get_overlapping_usernames(source_user_df, target_user_df, mapping_file_path)
    process_overlapping_usernames(conn_target, target_user_df, overlapping_usernames, overwrite_policy)
    cloned_users = create_users(conn_target, source_user_df, mapping_file_path, server_type)
    cloned_users = update_users(conn_target, source_user_df, mapping_file_path)
    # if server_type == 'tableau_server':
    #     cloned_users = update_users(conn_target, source_user_df, mapping_file_path)
    return cloned_users



from tableau_api_lib.api_requests import BaseRequest


class CreateGroupRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests creating groups.
    :param class ts_connection: the Tableau Server connection object
    :param str new_group_name: the name of the group being created
    :param str active_directory_group_name: the active directory group name
    :param str active_directory_domain_name: the active directory domain name
    :param str default_site_role: the default site role for the created group
    """
    def __init__(self,
                 ts_connection,
                 new_group_name=None,
                 active_directory_group_name=None,
                 active_directory_domain_name=None,
                 default_site_role=None):

        super().__init__(ts_connection)
        self._new_group_name = new_group_name
        self._active_directory_group_name = active_directory_group_name
        self._active_directory_domain_name = active_directory_domain_name
        self._default_site_role = default_site_role
        self.base_create_group_request()

    @property
    def required_group_param_keys(self):
        return ['name']

    @property
    def optional_source_param_keys(self):
        return [
            'source',
            'domainName',
            'siteRole'
        ]

    @property
    def required_group_param_values(self):
        return [self._new_group_name]

    @property
    def optional_source_param_values(self):
        return [
            self._active_directory_group_name,
            self._active_directory_domain_name,
            self._default_site_role
        ]

    def base_create_group_request(self):
        if all(self.required_group_param_values):
            self._request_body.update({'group': {}})
            self._request_body['group'].update(
                self._get_parameters_dict(self.required_group_param_keys,
                                          self.required_group_param_values))
        else:
            self._invalid_parameter_exception()
        return self._request_body

    def modified_create_group_request(self):
        if any(self.optional_source_param_values):
            self._request_body['group'].update({'import': {}})
            self._request_body['group']['import'].update(
                self._get_parameters_dict(self.optional_source_param_keys,
                                          self.optional_source_param_values))
        return self._request_body

    def get_request(self):
        return self.modified_create_group_request()



"""
Helper functions for querying REST API data for users
"""


import pandas as pd
from tableau_api_lib.utils import extract_pages


def get_all_user_fields(conn):
    all_users = extract_pages(conn.get_users_on_site, parameter_dict={'fields': 'fields=_all_'})
    return all_users


def get_all_user_names(conn):
    all_users = get_all_user_fields(conn)
    all_usernames = [user['name'] for user in all_users]
    return all_usernames


def get_all_user_emails(conn):
    all_users = get_all_user_fields(conn)
    all_user_emails = [user['email'] for user in all_users]
    return all_user_emails


def get_all_user_fullnames(conn):
    all_users = get_all_user_fields(conn)
    all_user_fullnames = [user['fullName'] for user in all_users]
    return all_user_fullnames


def get_all_user_roles(conn):
    all_users = get_all_user_fields(conn)
    all_user_roles = [user['siteRole'] for user in all_users]
    return all_user_roles


def get_users_dataframe(conn):
    users_df = pd.DataFrame(get_all_user_fields(conn))
    return users_df



"""
Functions to add:
get_favorites_dataframe
"""


import pandas as pd
from tableau_api_lib.utils import extract_pages


def get_all_favorite_fields(conn, user_id) -> list:
    """
    Queries all available favorites for the specified user on Tableau Server.
    :param class conn: the Tableau Server connection
    :param str user_id: the Tableau Server user whose favorites will be queried
    :return: list
    """
    all_favorites = conn.get_favorites_for_user(user_id).json()['favorites']['favorite']
    return all_favorites


def get_user_favorites_dataframe(conn, user_id) -> pd.DataFrame:
    """
    Returns a Pandas DataFrame containing details for the specified user's favorites on Tableau Server.
    :param class conn: the Tableau Server connection
    :param str user_id: the Tableau Server user whose favorites will be queried
    :return: pd.DataFrame
    """
    favorites_df = pd.DataFrame(get_all_favorite_fields(conn, user_id))
    return favorites_df



from tableau_api_lib.api_endpoints import BaseEndpoint


class FavoritesEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 add_to_favorites=False,
                 delete_from_favorites=False,
                 object_type=None,
                 object_id=None,
                 get_user_favorites=False,
                 user_id=None,
                 parameter_dict=None):
        """
        Builds the API endpoints for Favorites methods.
        :param class ts_connection: the Tableau Server connection object.
        :param bool add_to_favorites: True if adding item to favorites, False otherwise.
        :param bool delete_from_favorites: True if deleting item from favorites, False otherwise.
        :param str object_type: the Tableau object type being considered.
        :param str object_id: the ID of the specific Tableau object being considered.
        :param bool get_user_favorites: True if getting the user's favorite items, False otherwise.
        :param str user_id: the user ID.
        :param dict parameter_dict: dictionary of URL parameters to append. The value in each key-value pair is the literal
        text that will be appended to the URL endpoint.
        """

        super().__init__(ts_connection)
        self._add_to_favorites = add_to_favorites
        self._delete_from_favorites = delete_from_favorites
        self._object_type = object_type.lower() if object_type else None
        self._object_id = object_id
        self._get_user_favorites = get_user_favorites
        self._user_id = user_id
        self._parameter_dict = parameter_dict

    @property
    def valid_object_types(self):
        return [
            'datasource',
            'project',
            'workbook',
            'view'
        ]

    def _validate_object_type(self):
        if self._object_type.capitalize() in self.valid_object_types:
            pass
        else:
            self._invalid_parameter_exception()

    @property
    def base_favorites_url(self):
        return "{0}/api/{1}/sites/{2}/favorites".format(self._connection.server,
                                                        self._connection.api_version,
                                                        self._connection.site_id)

    @property
    def base_favorites_user_id_url(self):
        return "{0}/{1}".format(self.base_favorites_url,
                                self._user_id)

    @property
    def base_favorites_user_object_url(self):
        return "{0}/{1}s/{2}".format(self.base_favorites_user_id_url,
                                     self._object_type,
                                     self._object_id)

    def get_endpoint(self):
        if self._add_to_favorites and self._user_id:
            url = self.base_favorites_user_id_url
        elif self._delete_from_favorites and self._user_id:
            url = self.base_favorites_user_object_url
        elif self._get_user_favorites and self._user_id:
            url = self.base_favorites_user_id_url
        else:
            self._invalid_parameter_exception()

        return self._append_url_parameters(url)



from tableau_api_lib.api_endpoints import BaseEndpoint


class JobsEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 query_jobs=False,
                 query_job=False,
                 cancel_job=False,
                 job_id=None,
                 parameter_dict=None):
        """
        Builds API endpoints for REST API job methods.
        :param class ts_connection: the Tableau Server connection object
        :param bool query_jobs: True if querying all jobs, False otherwise
        :param bool query_job: True if querying a specific job, False otherwise
        :param bool cancel_job: True if canceling a specific job, False otherwise
        :param str job_id: the job ID
        :param dict parameter_dict: dictionary of URL parameters to append; the value in each key-value pair is the
        literal text that will be appended to the URL endpoint
        """

        super().__init__(ts_connection)
        self._query_jobs = query_jobs
        self._query_job = query_job
        self._cancel_job = cancel_job
        self._job_id = job_id
        self._parameter_dict = parameter_dict

    @property
    def mutually_exclusive_params(self):
        return [
            self._query_jobs,
            self._query_job,
            self._cancel_job
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_job_url(self):
        return "{0}/api/{1}/sites/{2}/jobs".format(self._connection.server,
                                                   self._connection.api_version,
                                                   self._connection.site_id)

    @property
    def base_job_id_url(self):
        return "{0}/{1}".format(self.base_job_url,
                                self._job_id)

    def get_endpoint(self):
        if self._job_id:
            if self._query_job and not self._cancel_job:
                url = self.base_job_id_url
            elif self._cancel_job and not self._query_job:
                url = self.base_job_id_url
            else:
                url = self._invalid_parameter_exception()
        else:
            url = self.base_job_url

        return self._append_url_parameters(url)



from tableau_api_lib.exceptions import InvalidParameterException


class BaseEndpoint:
    def __init__(self,
                 ts_connection):
        """
        The base endpoint from which all other API endpoints are built.
        :param class ts_connection: the Tableau Server connection object
        """

        self._connection = ts_connection
        self._parameter_dict = {}
        
    @property
    def _params_exist(self):
        if self._parameter_dict:
            return list(self._parameter_dict.keys())
        else:
            return []
        
    @property
    def _params_text(self):
        if self._parameter_dict:
            return list(self._parameter_dict.values())
        else:
            return []

    @property
    def base_url(self):
        return self._connection.base_url
    
    def _append_url_parameters(self, url):
        text_to_append = "?" if any(self._params_exist) else ""
        for i, text in enumerate(self._params_text):
            if self._params_exist[i]:
                text_to_append += text if text_to_append.endswith('?') else ('&' + text)
        return "{0}{1}".format(url, text_to_append)
    
    def _invalid_parameter_exception(self):
        raise InvalidParameterException(self.__class__.__name__,
                                        vars(self))



from .projects import clone_projects
from .sites import clone_sites
from .schedules import clone_schedules, clone_schedule_state
from .users import clone_users
from .groups import clone_groups
from .datasources import clone_datasources



from tableau_api_lib.tableau_server_connection import TableauServerConnection
from tableau_api_lib.sample import sample_config
from tableau_api_lib import decorators
from tableau_api_lib import api_endpoints
from tableau_api_lib import exceptions
from tableau_api_lib import api_requests
from tableau_api_lib import utils

name = 'tableau_api_lib'



"""
Helper functions for querying REST API data for workbooks and views
"""


import pandas as pd
from tableau_api_lib.utils import extract_pages, flatten_dict_column
from tableau_api_lib.utils.querying.sites import get_active_site_id
from tableau_api_lib.exceptions.tableau_server_exceptions import ContentNotFound


def get_all_workbook_fields(conn):
    all_workbooks = extract_pages(conn.query_workbooks_for_site, parameter_dict={'fields': 'fields=_all_'})
    return all_workbooks


def get_workbooks_dataframe(conn):
    workbooks_df = pd.DataFrame(get_all_workbook_fields(conn))
    return workbooks_df


def get_all_view_fields(conn, site_id):
    all_views = extract_pages(conn.query_views_for_site, content_id=site_id, parameter_dict={'fields': 'fields=_all_'})
    return all_views


def get_views_dataframe(conn, site_id=None):
    if not site_id:
        site_id = get_active_site_id(conn)
    views_df = pd.DataFrame(get_all_view_fields(conn, site_id))
    views_df = flatten_dict_column(views_df, keys=['totalViewCount'], col_name='usage')
    return views_df


def extract_datasource_details(df, keys):
    for key in keys:
        df['datasource_' + key] = df['datasource'].apply(lambda project: project[key])
    df.drop(columns=['datasource'], inplace=True)
    return df


def get_workbook_connections_dataframe(conn, workbook_id):
    try:
        connections_json = conn.query_workbook_connections(workbook_id).json()['connections']['connection']
        connections_df = pd.DataFrame(connections_json)
        connections_df = flatten_dict_column(connections_df, keys=['id', 'name'], col_name='datasource')
        return connections_df
    except KeyError:
        raise ContentNotFound('workbook', workbook_id)


def get_embedded_datasources_dataframe(conn, workbooks_df, workbook_ids=None):
    """
    Creates a Pandas DataFrame of all embedded workbook datasources, or specific workbooks if specified.
    :param TableauServerConnection conn: the Tableau Server connection
    :param pd.DataFrame workbooks_df: the workbook DataFrame containing details for all workbooks
    :param list workbook_ids: a list of workbook IDs whose embedded datasources will be queried
    :return:
    """
    workbook_ids = workbook_ids or []
    workbook_ids = workbook_ids.to_list() if isinstance(workbook_ids, pd.core.series.Series) else workbook_ids
    if any(workbook_ids):
        workbooks_df = workbooks_df[workbooks_df['id'].isin(workbook_ids)]
    embedded_datasources_df = pd.DataFrame()
    for index, workbook in workbooks_df.iterrows():
        workbook_connections_df = get_workbook_connections_dataframe(conn=conn, workbook_id=workbook['id'])
        embedded_datasources_df = embedded_datasources_df.append(workbook_connections_df, ignore_index=True, sort=True)
    return embedded_datasources_df



from tableau_api_lib.api_requests import BaseRequest


class CreateScheduleRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests creating schedules.
    :param class ts_connection: the Tableau Server connection object
    :param str schedule_name: the name of the schedule being created
    :param str schedule_priority: the priority value (1-100) for the schedule
    :param str schedule_type: the schedule type (Flow, Extract, or Subscription)
    :param str schedule_execution_order: set this value to 'Parallel' to allow jobs associated with this schedule to run
    in parallel; set the value to 'Serial' to require the jobs to run one at a time
    :param str schedule_frequency: the granularity of the schedule (Hourly, Daily, Weekly, or Monthly)
    :param str start_time: the time of day when the schedule should run (HH:MM:SS). If the frequency is set to 'Hourly',
    this value indicates the hour the schedule starts running
    :param str end_time: only set this value if the schedule frequency has been set to 'Hourly'. This value indicates
    the hour the schedule will stop running (HH:MM:SS)
    :param list interval_expression_list: this list of dicts specifies the time interval(s) between jobs on the
    schedule. The value required here depends on the 'schedule_frequency' value. If 'schedule_frequency' = 'Hourly',
    the interval expression should be hours="interval" (where "interval" is a number [1, 2, 4, 6, 8, 12] in quotes).
    If 'schedule_frequency' = 'Daily', no interval needs to be specified.
    If 'schedule_frequency' = 'Weekly, the interval is weekDay="weekday", where weekday is one of
    ['Sunday', 'Monday', 'Tuesday', etc.] wrapped in quotes.
    If 'schedule_frequency' = 'Monthly', the interval expression is monthDay="day", where day is either the day of the
    month (1-31), or 'LastDay'. In both cases the value is wrapped in quotes.
    """
    def __init__(self,
                 ts_connection,
                 schedule_name,
                 schedule_priority=50,
                 schedule_type='Extract',
                 schedule_execution_order='Parallel',
                 schedule_frequency='Weekly',
                 start_time='07:00:00',
                 end_time=None,
                 interval_expression_list=[{'weekDay': 'Monday'}]
                 ):

        super().__init__(ts_connection)
        self._schedule_name = schedule_name
        self._schedule_priority = schedule_priority
        self._schedule_type = schedule_type
        self._schedule_execution_order = schedule_execution_order
        self._schedule_frequency = schedule_frequency
        self._start_time = start_time
        self._end_time = end_time
        self._interval_expression_list = interval_expression_list
        self._interval_expression_keys, self._interval_expression_values = None, None
        self._validate_inputs()
        self.base_create_schedule_request()

    @property
    def required_schedule_param_keys(self):
        return [
            'name',
            'priority',
            'type',
            'frequency',
            'executionOrder'
        ]

    @property
    def required_frequency_param_keys(self):
        return [
            'start',
            'end'
        ]

    @property
    def valid_interval_keys(self):
        return [
            'minutes',
            'hours',
            'weekDay',
            'monthDay'
        ]

    def _validate_inputs(self):
        valid = True
        if self._interval_expression_list:
            self._set_interval_expressions()
        else:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def required_schedule_param_values(self):
        return [
            self._schedule_name,
            self._schedule_priority,
            self._schedule_type,
            self._schedule_frequency,
            self._schedule_execution_order
        ]

    @property
    def required_frequency_param_values(self):
        return [
            self._start_time,
            self._end_time
        ]

    def _unpack_interval_expressions_list(self, interval_list):
        interval_keys = []
        interval_values = []
        for interval_dict in interval_list:
            for key in interval_dict.keys():
                if key in self.valid_interval_keys:
                    interval_keys.append(key)
                    interval_values.append(interval_dict[key])
                else:
                    self._invalid_parameter_exception()
        return interval_keys, interval_values

    def _set_interval_expressions(self):
        if self._interval_expression_list:
            if any(self._interval_expression_list):
                self._interval_expression_keys, self._interval_expression_values = \
                    self._unpack_interval_expressions_list(
                        self._interval_expression_list)

    @staticmethod
    def _get_parameters_list(param_keys, param_values):
        params_list = []
        for i, key in enumerate(param_keys):
            if param_values[i]:
                params_list.append({key: param_values[i]})
        return params_list

    def base_create_schedule_request(self):
        self._request_body.update({'schedule': {'frequencyDetails': {}}})
        return self._request_body

    def modified_create_schedule_request(self):
        self._request_body['schedule'].update(
            self._get_parameters_dict(self.required_schedule_param_keys,
                                      self.required_schedule_param_values))
        self._request_body['schedule']['frequencyDetails'].update(
            self._get_parameters_dict(self.required_frequency_param_keys,
                                      self.required_frequency_param_values))
        if self._interval_expression_list:
            self._request_body['schedule']['frequencyDetails'].update({'intervals': {}})
            self._request_body['schedule']['frequencyDetails']['intervals'].update(
                {'interval': self._get_parameters_list(self._interval_expression_keys,
                                                       self._interval_expression_values)})
        return self._request_body

    def get_request(self):
        return self.modified_create_schedule_request()



from tableau_api_lib.api_requests import BaseRequest


class CreateProjectRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests creating projects.
    :param class ts_connection: the Tableau Server connection object
    :param str project_name: the name of the project being created
    :param str project_description: the description of the project being created
    :param str parent_project_id: the project ID for the parent project, if creating a child project
    """
    def __init__(self,
                 ts_connection,
                 project_name,
                 project_description=None,
                 content_permissions='ManagedByOwner',
                 parent_project_id=None):
        super().__init__(ts_connection)
        self._project_name = project_name
        self._project_description = project_description
        self._content_permissions = content_permissions
        self._parent_project_id = parent_project_id
        self.base_create_project_request()

    @property
    def optional_project_param_keys(self):
        return [
            'parentProjectId',
            'description',
            'contentPermissions'
        ]

    @property
    def optional_project_param_values(self):
        return [
            self._parent_project_id,
            self._project_description,
            self._content_permissions
        ]

    def base_create_project_request(self):
        self._request_body.update({'project': {'name': self._project_name}})
        return self._request_body

    def modified_create_project_request(self):
        self._request_body['project'].update(
            self._get_parameters_dict(
                self.optional_project_param_keys,
                self.optional_project_param_values))
        return self._request_body

    def get_request(self):
        return self.modified_create_project_request()



from tableau_api_lib.api_requests import BaseRequest


class UpdateDataAlertRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests updating database columns.
    :param class ts_connection: the Tableau Server connection object
    :param str data_alert_subject: (optional) the string to set as the new subject of the alert
    :param str data_alert_frequency: (optional) the frequency of the data-driven alert: once, frequently, hourly, daily,
    or weekly
    :param str data_alert_owner_id: (optional) the ID of the user to assign as owner of the data-driven alert
    :param bool is_public_flag: (optional) Determines the visibility of the data-driven alert. If the flag is
    True, users with access to the view containing the alert can see the alert and add themselves as recipients. If the
    flag is False, then the alert is only visible to the owner, site or server administrators, and specific users they
    add as recipients
    """
    def __init__(self,
                 ts_connection,
                 data_alert_subject=None,
                 data_alert_frequency=None,
                 data_alert_owner_id=None,
                 is_public_flag=None):
        super().__init__(ts_connection)
        self._data_alert_subject = data_alert_subject
        self._data_alert_frequency = data_alert_frequency
        self._data_alert_owner_id = data_alert_owner_id
        self._is_public_flag = is_public_flag
        self.base_update_alert_request()

    @property
    def optional_alert_param_keys(self):
        return [
            'subject',
            'frequency',
            'public'
        ]

    @property
    def optional_owner_param_keys(self):
        return ['id']

    @property
    def optional_alert_param_values(self):
        return [
            self._data_alert_subject,
            self._data_alert_frequency,
            self._is_public_flag
        ]

    @property
    def optional_owner_param_values(self):
        return [self._data_alert_owner_id]

    def base_update_alert_request(self):
        self._request_body.update({'dataAlert': {}})
        return self._request_body

    def modified_update_alert_request(self):
        self._request_body['dataAlert'].update(
            self._get_parameters_dict(
                self.optional_alert_param_keys,
                self.optional_alert_param_values))
        if self._data_alert_owner_id:
            self._request_body['dataAlert'].update({'owner': {}})
            self._request_body['dataAlert']['owner'].update(
                self._get_parameters_dict(
                    self.optional_owner_param_keys,
                    self.optional_owner_param_values))
        return self._request_body

    def get_request(self):
        return self.modified_update_alert_request()



from tableau_api_lib.api_endpoints import BaseEndpoint


class FileUploadEndpoint(BaseEndpoint):
    """
    FileUploadEndpoint endpoint for Tableau Server API api_requests.

    :param ts_connection:           The Tableau Server connection object.
    :type ts_connection:            class
    :param initiate_file_upload:    Boolean flag; True if initiating a file upload, False otherwise.
    :type initiate_file_upload:     boolean
    :param append_to_file_upload:   Boolean flag; True if appending to a file upload, False otherwise.
    :type append_to_file_upload:    boolean
    :param upload_session_id:       The upload session ID.
    :type upload_session_id:        string
    :param parameter_dict:          Dictionary of URL parameters to append. The value in each key-value pair
                                    is the literal text that will be appended to the URL endpoint.
    :type parameter_dict:           dict
    """
    def __init__(self,
                 ts_connection,
                 initiate_file_upload=False,
                 append_to_file_upload=False,
                 upload_session_id=None,
                 parameter_dict=None):

        super().__init__(ts_connection)
        self._initiate_file_upload = initiate_file_upload
        self._append_to_file_upload = append_to_file_upload
        self._upload_session_id = upload_session_id
        self._parameter_dict = parameter_dict

    @property
    def mutually_exclusive_params(self):
        return [
            self._initiate_file_upload,
            self._append_to_file_upload
        ]

    def _validate_inputs(self):
        valid = True
        if sum(self.mutually_exclusive_params) != 1:
            valid = False
        if not valid:
            self._invalid_parameter_exception()

    @property
    def base_file_upload_url(self):
        return "{0}/api/{1}/sites/{2}/fileUploads".format(self._connection.server,
                                                          self._connection.api_version,
                                                          self._connection.site_id)

    @property
    def base_file_upload_id_url(self):
        return "{0}/{1}".format(self.base_file_upload_url,
                                self._upload_session_id)

    def get_endpoint(self):
        url = None
        if self._initiate_file_upload:
            url = self.base_file_upload_url
        elif self._append_to_file_upload and self._upload_session_id:
            url = self.base_file_upload_id_url
        else:
            self._invalid_parameter_exception()

        return self._append_url_parameters(url)



from tableau_api_lib.exceptions import ContentNotFound, PaginationError


def get_page_attributes(query, query_func):
    """
    Get page attributes (pageNumber, pageSize, totalAvailable) from a REST API paginated response.
    :param dict query: the results of the GET request query, containing paginated data
    :param function query_func: a callable function that will issue a GET request to Tableau Server
    :return: page_number, page_size, total_available
    """
    try:
        pagination = query['pagination']
        page_number = int(pagination['pageNumber'])
        page_size = int(pagination['pageSize'])
        total_available = int(pagination['totalAvailable'])
        return page_number, page_size, total_available
    except KeyError:
        raise PaginationError(query_func)


def extract_pages(query_func,
                  content_id=None,
                  *,
                  starting_page=1,
                  page_size=100,
                  limit=None,
                  parameter_dict=None):
    """
    Extracts pages from paginated Tableau Server API responses.
    :param function query_func: a callable function that will issue a GET request to Tableau Server
    :param str content_id: the content ID value, if the query_func requires one [group_id, site_id, etc.]
    :param int starting_page: the page number to start on. Defaults to the first page (page_number = 1)
    :param int page_size: the number of objects per page. If querying users, this is the number of users per page
    :param int limit: the maximum number of objects to return. Default is no limit
    :param dict parameter_dict: a dict whose values are appended to the REST API URL endpoint as URL parameters
    :return: JSON or dict
    """
    parameter_dict = parameter_dict or {}
    extracted_pages = []
    page_number = starting_page
    extracting = True

    while extracting:
        parameter_dict.update({
            'pageNumber': 'pageNumber={}'.format(page_number),
            'pageSize': 'pageSize={}'.format(page_size)
        })
        query = process_query(query_func, content_id, parameter_dict)
        page_number, page_size, total_available = get_page_attributes(query, query_func)
        if total_available == 0:
            raise ContentNotFound()
        extracted_pages, extracting, page_number = update_pagination_params(query,
                                                                            extracted_pages,
                                                                            page_number,
                                                                            page_size,
                                                                            total_available,
                                                                            limit,
                                                                            extracting)
    return extracted_pages


def process_query(query_func, content_id, parameter_dict):
    if content_id:
        if isinstance(content_id, str):
            try:
                query = query_func(content_id, parameter_dict=parameter_dict).json()
            except TypeError:
                raise PaginationError(query_func)
        else:
            raise TypeError('The content ID value passed to {} must be a string.'.format(query_func.__name__))
    else:
        try:
            query = query_func(parameter_dict=parameter_dict).json()
        except TypeError:
            raise PaginationError(query_func)
    return query


def update_pagination_params(query, extracted_pages, page_number, page_size, total_available, limit, extracting):
    try:
        outer_key = [key for key in query.keys() if key != 'pagination'].pop()
        inner_key = list(query[outer_key].keys()).pop()
        extracted_pages += query[outer_key][inner_key]
    except IndexError:
        raise ContentNotFound()

    if limit:
        if limit <= len(extracted_pages):
            extracted_pages = extracted_pages[:limit]
            extracting = False
    elif total_available <= (page_number * page_size):
        extracting = False
    else:
        page_number += 1
    return extracted_pages, extracting, page_number



from tableau_api_lib.api_requests import BaseRequest


class AddUserToGroupRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests adding users to groups.
    :param class ts_connection: the Tableau Server connection object
    :param str user_id: the user ID
    """
    def __init__(self,
                 ts_connection,
                 user_id):
        super().__init__(ts_connection)
        self._user_id = user_id
        self.base_add_user_request()

    @property
    def required_user_param_keys(self):
        return ['id']

    @property
    def required_user_param_values(self):
        return [self._user_id]

    def base_add_user_request(self):
        self._request_body.update({'user': {}})
        return self._request_body

    def modified_add_user_request(self):
        self._request_body['user'].update(
            self._get_parameters_dict(self.required_user_param_keys,
                                      self.required_user_param_values))
        return self._request_body

    def get_request(self):
        return self.modified_add_user_request()



from tableau_api_lib.api_requests import BaseRequest


class CreateSubscriptionRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests creating subscriptions.
    :param class ts_connection: the Tableau Server connection object
    :param str subscription_subject: the subject to display to users receiving the subscription
    :param str content_type: set this value to 'Workbook' if the subscription is for a workbook; set this value to
    'View' if the subscription is for a view
    :param str content_id: the ID of the workbook or view the subscription is sourced from.
    :param str schedule_id: the ID of the schedule the subscription runs on.
    :param str user_id: the user ID for the user who is being subscribed to the view or workbook; this user must have
    an email address defined on Tableau Server.
    :param bool attach_image_flag: True if attaching a .png image to the subscription, defaults to False
    :param bool attach_pdf_flag: True if attaching a .pdf file to the subscription, defaults to False
    """
    def __init__(self,
                 ts_connection,
                 subscription_subject,
                 content_type,
                 content_id,
                 schedule_id,
                 user_id,
                 attach_image_flag=None,
                 attach_pdf_flag=None):

        super().__init__(ts_connection)
        self._subscription_subject = subscription_subject
        self._content_type = content_type.lower()
        self._content_id = content_id
        self._schedule_id = schedule_id
        self._user_id = user_id
        self._attach_image_flag = attach_image_flag
        self._attach_pdf_flag = attach_pdf_flag
        self._validate_content_type()

    @property
    def valid_content_types(self):
        return [
            'Workbook',
            'View'
        ]

    def _validate_content_type(self):
        valid = True
        if not(self._content_type.lower().capitalize() in self.valid_content_types):
            valid = True
        if not valid:
            self._invalid_parameter_exception()

    @property
    def optional_param_keys(self):
        return [
            'attachImage',
            'attachPdf'
        ]

    @property
    def optional_param_values(self):
        return [
            self._attach_image_flag,
            self._attach_pdf_flag
        ]

    def base_create_subscription_request(self):
        self._request_body.update({
            'subscription': {
                'subject': self._subscription_subject,
                'content': {
                    'type': self._content_type,
                    'id': self._content_id
                },
                'schedule': {'id': self._schedule_id},
                'user': {'id': self._user_id}
            },
        })
        if any(self.optional_param_values):
            self._request_body['subscription'].update(self._get_parameters_dict(self.optional_param_keys,
                                                                                self.optional_param_values))
        return self._request_body

    def get_request(self):
        return self.base_create_subscription_request()



"""
Helper functions for querying REST API data for flows
"""


import pandas as pd
from tableau_api_lib.utils import extract_pages


def get_all_flow_fields(conn) -> list:
    """
    Queries all available flow fields from Tableau Server.
    :param class conn: the Tableau Server connection
    :return: list
    """
    all_flows = extract_pages(conn.query_flows_for_site, parameter_dict={'fields': 'fields=_all_'})
    return all_flows


def get_flows_dataframe(conn) -> pd.DataFrame:
    """
    Returns a Pandas DataFrame of all available Tableau Server flow fields.
    :param class conn: the Tableau Server connection
    :return: pd.DataFrame
    """
    flows_df = pd.DataFrame(get_all_flow_fields(conn))
    return flows_df



import os
import json
from urllib3.fields import RequestField
from urllib3.filepost import encode_multipart_formdata

from tableau_api_lib.api_requests import BaseRequest
from tableau_api_lib.exceptions import InvalidFileTypeException


CHUNK_SIZE = 1024 * 1024 * 5  # 5MB
FILE_SIZE_LIMIT = 1024 * 1024 * 60  # 60MB


class PublishWorkbookRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests publishing workbooks.
    :param class ts_connection: the Tableau Server connection object
    :param string workbook_name: the name the published workbook will have on Tableau Server
    :param string workbook_file_path: the file path for the workbook being published
    :param string project_id: the project ID of the project the workbook is being published to
    :param boolean show_tabs_flag: (Optional) true if the workbook will show views as tabs, false otherwise
    :param string user_id: If generating thumbnails as a specific user, specify the user ID here
    :param list server_address: (Optional) specify the server addresses for data source connections if the data sources
    do not use OAuth
    :param list port_number: (Optional) specify the port numbers for a data source connections if that data sources do
    not use OAuth
    :param list connection_username: (Optional) if the workbook's data source connections require credentials, the
    <connectionCredentials> elements are included and this attribute specifies the connection username. If the element
    is included but is not required (for example, if the data source uses OAuth), the server ignores the element and its
    attributes
    :param list connection_password: (Optional) if the workbook's data source connections require credentials, the
    <connectionCredentials> elements are included and this attribute specifies the connection password. If the element
    is included but is not required (for example, if the data source uses OAuth), the server ignores the element and its
    attributes
    :param list embed_credentials_flag: (Optional) true if embedding credentials in the published workbook, false
    otherwise
    :param list oauth_flag: list of boolean flags; True if OAuth is used for the credentials, false otherwise.
    :param list workbook_views_to_hide: a list of the views to hide for the workbook being published. The list should
    contain the view names, not view IDs
    :param boolean hide_view_flag: (Optional) true if the published workbook will hide any of its views,  false
    otherwise
    """
    def __init__(self,
                 ts_connection,
                 workbook_name,
                 workbook_file_path,
                 project_id,
                 show_tabs_flag=False,
                 user_id=None,
                 server_address=None,
                 port_number=None,
                 connection_username=None,
                 connection_password=None,
                 embed_credentials_flag=False,
                 oauth_flag=False,
                 workbook_views_to_hide=None,
                 hide_view_flag=False):

        super().__init__(ts_connection)
        self._workbook_name = workbook_name
        self._workbook_file_path = workbook_file_path
        self._project_id = project_id
        self._show_tabs_flag = show_tabs_flag
        self._user_id = user_id
        self._server_address = server_address
        self._port_number = port_number
        self._connection_username = connection_username
        self._connection_password = connection_password
        self._embed_credentials_flag = embed_credentials_flag
        self._oauth_flag = oauth_flag
        self._workbook_views_to_hide = workbook_views_to_hide
        self._hide_view_flag = hide_view_flag
        self.payload = None
        self.content_type = None
        self._verify_embed_requirements()
        self._listify_inputs()
        self._file_is_chunked = self._file_requires_chunking()
        self.base_publish_workbook_request()

    def _listify_inputs(self):
        if any(self.optional_connection_param_values):
            if isinstance(self._server_address, str):
                self._server_address = [self._server_address]
            if isinstance(self._port_number, str):
                self._port_number = [self._port_number]
        if self._connection_username:
            if isinstance(self._connection_username, str):
                self._connection_username = [self._connection_username]
        if self._connection_password:
            if isinstance(self._connection_password, str):
                self._connection_password = [self._connection_password]
        if self._embed_credentials_flag:
            if isinstance(self._embed_credentials_flag, bool) or isinstance(self._embed_credentials_flag, str):
                self._embed_credentials_flag = [self._embed_credentials_flag]
        if isinstance(self._oauth_flag, bool) or isinstance(self._oauth_flag, str):
            self._oauth_flag = [self._oauth_flag]

    def _verify_embed_requirements(self):
        if self._embed_credentials_flag:
            if self._server_address and self._connection_username and self._connection_password:
                pass
            else:
                self._invalid_parameter_exception()

    @property
    def valid_file_extensions(self):
        return [
            'twb',
            'twbx'
        ]

    @property
    def optional_workbook_param_keys(self):
        return [
            'showTabs',
            'generateThumbnailsAsUser'
        ]

    @property
    def optional_connection_param_keys(self):
        return [
            'serverAddress',
            'serverPort'
        ]

    @property
    def optional_credentials_param_keys(self):
        return [
            'name',
            'password',
            'embed',
            'oAuth'
        ]

    @property
    def optional_view_param_keys(self):
        return [
            'name',
            'hidden'
        ]

    @property
    def optional_workbook_param_values(self):
        return [
            self._show_tabs_flag,
            self._user_id
        ]

    @property
    def optional_connection_param_values(self):
        return [
            self._server_address,
            self._port_number
        ]

    @property
    def optional_credentials_param_values(self):
        return [
            self._connection_username,
            self._connection_password,
            self._embed_credentials_flag,
            self._oauth_flag
        ]

    @property
    def optional_view_param_values(self):
        values_list = []
        if self._hide_view_flag:
            [values_list.append(view_name) for view_name in self._workbook_views_to_hide]
        else:
            [values_list.append(None) for _ in self.optional_view_param_keys]
        return values_list

    def base_publish_workbook_request(self):
        self._request_body.update({
            'workbook': {
                'name': self._workbook_name,
                'project': {'id': self._project_id}
            }
        })
        return self._request_body

    def modified_publish_workbook_request(self):
        self._request_body['workbook'].update(self._get_parameters_dict(self.optional_workbook_param_keys,
                                                                        self.optional_workbook_param_values))
        if any(self.optional_connection_param_values) or any(self.optional_credentials_param_values):
            self._request_body['workbook'].update({'connections': {'connection': []}})
            for i, _ in enumerate(self._connection_username or [None]):
                self._request_body['workbook']['connections']['connection'].append({
                    'serverAddress': self._server_address[i] if self._server_address else None,
                    'serverPort': self._port_number[i] if self._port_number else None,
                    'connectionCredentials': {
                        'name': self._connection_username[i] if self._connection_username else None,
                        'password': self._connection_password[i] if self._connection_password else None,
                        'embed': self._embed_credentials_flag[i] if self._embed_credentials_flag else False
                    }
                })
            if any(self._oauth_flag):
                for i, _ in enumerate(self._oauth_flag):
                    self._request_body['workbook']['connections']['connection'][i]['connectionCredentials'].update({
                        'oAuth': self._oauth_flag[i]
                    })

        if any(self.optional_view_param_values):
            self._request_body['workbook'].update({'views': {'view': []}})
            for i, view in enumerate(self.optional_view_param_values):
                self._request_body['workbook']['views']['view'].append({
                    'name': view,
                    'hidden': 'true'
                })
        return self._request_body

    def _file_requires_chunking(self):
        file_size = os.path.getsize(self._workbook_file_path)
        if file_size > FILE_SIZE_LIMIT:
            return True

    def get_workbook(self):
        workbook_file = os.path.basename(self._workbook_file_path)
        with open(self._workbook_file_path, 'rb') as f:
            workbook_bytes = f.read()
        file_extension = workbook_file.split('.')[-1]
        if file_extension in self.valid_file_extensions:
            pass
        else:
            raise InvalidFileTypeException(self.__class__.__name__,
                                           file_variety='workbook',
                                           file_extension=file_extension)
        return workbook_file, workbook_bytes

    # testing for chunk upload
    def publish_prep(self, publish_content_type, parameter_dict):
        parameter_dict = parameter_dict if parameter_dict else {'overwrite': 'overwrite=true'}
        filename = os.path.basename(self._workbook_file_path)
        file_extension = filename.split('.')[-1]

        if self._file_is_chunked:
            upload_session_id = self._connection.initiate_file_upload().json()['fileUpload']['uploadSessionId']
            parameter_dict.update({'param': 'uploadSessionId={}'.format(upload_session_id)})
            chunks = self.read_chunks(self._workbook_file_path)
            for chunk in chunks:
                request, append_content_type = self.chunk_req(chunk)
                file_upload = self._connection.append_to_file_upload(upload_session_id=upload_session_id,
                                                                     payload=request,
                                                                     content_type=append_content_type)

        publishing_headers = self._connection.default_headers.copy()
        publishing_headers.update({'content-type': publish_content_type})
        parameter_dict.update({'workbookType': 'workbookType={}'.format(file_extension)})
        return publishing_headers, parameter_dict

    @staticmethod
    def read_chunks(file_path):
        with open(file_path, 'rb') as f:
            while True:
                chunked_content = f.read(CHUNK_SIZE)
                if not chunked_content:
                    break
                yield chunked_content

    def chunk_req(self, chunk):
        parts = {'request_payload': (None, '', 'application/json'),
                 'tableau_file': ('file', chunk, 'application/octet-stream')}
        return self._add_multipart(parts)

    @staticmethod
    def _add_multipart(parts):
        mime_multipart_parts = list()
        for name, (filename, data, content_type) in parts.items():
            multipart_part = RequestField(name=name, data=data, filename=filename)
            multipart_part.make_multipart(content_type=content_type)
            mime_multipart_parts.append(multipart_part)
        request, content_type = encode_multipart_formdata(mime_multipart_parts)
        content_type = ''.join(('multipart/mixed',) + content_type.partition(';')[1:])
        return request, content_type

    def _publish_chunked_file_request(self):
        request = self.modified_publish_workbook_request()
        parts = {'request_payload': (None, json.dumps(request), 'application/json')}
        return self._add_multipart(parts)

    def _publish_single_file_request(self):
        request = self.modified_publish_workbook_request()
        workbook_file, workbook_bytes = self.get_workbook()
        parts = {'request_payload': (None, json.dumps(request), 'application/json'),
                 'tableau_workbook': (workbook_file, workbook_bytes, 'application/octet-stream')}
        return self._add_multipart(parts)

    def get_request(self):
        if self._file_is_chunked:
            return self._publish_chunked_file_request()
        else:
            return self._publish_single_file_request()



from tableau_api_lib.api_requests import BaseRequest


class GraphqlRequest(BaseRequest):
    """
    Builds the request body for Tableau Server Metadata API GraphQL queries.
    :param class ts_connection: the Tableau Server connection object
    :param str query: the GraphQL query for the Metadata API (raw text)
    """
    def __init__(self,
                 ts_connection,
                 query):

        super().__init__(ts_connection)
        self._query = query
        self._validate_inputs()
        self.base_create_group_request()

    def _validate_inputs(self):
        if isinstance(self._query, str):
            pass
        else:
            raise ValueError("The GraphQL query argument must be of type 'str', not '{}'.".format(type(self._query)))

    def base_create_group_request(self):
        self._request_body.update({'query': self._query})
        return self._request_body

    def get_request(self):
        return self.base_create_group_request()



from tableau_api_lib.api_endpoints import BaseEndpoint


class DatabaseEndpoint(BaseEndpoint):
    def __init__(self,
                 ts_connection,
                 query_database=False,
                 query_databases=False,
                 update_database=False,
                 remove_database=False,
                 database_id=None):
        """
        Builds the API endpoints for interacting with database assets on Tableau Server.
        :param ts_connection: the Tableau Server connection object
        :param query_database: True if querying a specific database, False otherwise
        :param query_databases: True if querying all databases on the active site, False otherwise
        :param update_database: True if updating details for a specific database, False otherwise
        :param remove_database: True if removing a specific database asset, False otherwise
        :param database_id: the database ID
        """

        super().__init__(ts_connection)
        self._query_database = query_database
        self._query_databases = query_databases
        self._update_database = update_database
        self._remove_database = remove_database
        self._database_id = database_id
        self._validate_inputs()

    @property
    def mutually_exclusive_params(self):
        return [
            self._query_database,
            self._query_databases,
            self._update_database,
            self._remove_database
        ]

    def _validate_inputs(self):
        if sum(self.mutually_exclusive_params) == 1:
            pass
        else:
            self._invalid_parameter_exception()

    @property
    def base_database_url(self):
        return "{0}/api/{1}/sites/{2}/databases".format(self._connection.server,
                                                        self._connection.api_version,
                                                        self._connection.site_id)

    @property
    def base_database_id_url(self):
        return "{0}/{1}".format(self.base_database_url,
                                self._database_id)

    def get_endpoint(self):
        if self._database_id:
            url = self.base_database_id_url
        else:
            url = self.base_database_url
        if url:
            return url
        else:
            self._invalid_parameter_exception()



from tableau_api_lib.api_requests import BaseRequest


class UpdateTableRequest(BaseRequest):
    def __init__(self,
                 ts_connection,
                 certification_status=None,
                 certification_note=None,
                 new_description_value=None,
                 new_contact_id=None
                 ):
        """
        Builds the request body for Tableau Server REST API requests updating database tables.
        :param class ts_connection: the Tableau Server connection object
        :param bool certification_status: certifies (True) or removes certification (False) for the specified table
        :param str certification_note: custom text to accompany the certification status
        :param str new_description_value: custom text describing the table
        :param str new_contact_id: the ID for the Tableau Server user who is the contact for the specified database
        """

        super().__init__(ts_connection)
        self._certification_status = certification_status
        self._certification_note = certification_note
        self._new_description_value = new_description_value
        self._new_contact_id = new_contact_id
        self._request_body = {'table': {}}

    @property
    def optional_param_keys(self):
        return [
            'isCertified',
            'certificationNote',
            'description'
        ]

    @property
    def optional_param_values(self):
        return [
            self._certification_status,
            self._certification_note,
            self._new_description_value
        ]

    def base_update_table_request(self):
        if self._new_contact_id:
            self._request_body['table'].update({'contact': self._new_contact_id})
        self._request_body['table'].update(self._get_parameters_dict(self.optional_param_keys,
                                                                     self.optional_param_values))
        return self._request_body

    def get_request(self):
        return self.base_update_table_request()



from tableau_api_lib.api_requests import BaseRequest


class UpdateDatasourceConnectionRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests updating datasource connections.
    :param class ts_connection: the Tableau Server connection object
    :param str server_address: the new server address for the connection
    :param str port: the new port for the connection
    :param str connection_username: the new username for the connection
    :param str connection_password: the new password for the connection
    :param bool embed_password_flag: True if embedding password in the datasource connection, False otherwise
    """
    def __init__(self,
                 ts_connection,
                 server_address=None,
                 port=None,
                 connection_username=None,
                 connection_password=None,
                 embed_password_flag=None):

        super().__init__(ts_connection)
        self._server_address = server_address
        self._port = port
        self._connection_username = connection_username
        self._connection_password = connection_password
        self._embed_password_flag = embed_password_flag
        self._validate_inputs()
        self.base_update_datasource_connection_request()

    @property
    def optional_parameter_keys(self):
        return [
            'serverAddress',
            'serverPort',
            'userName',
            'password',
            'embedPassword'
        ]

    @property
    def optional_parameter_values_exist(self):
        return [
            self._server_address,
            self._port,
            self._connection_username,
            self._connection_password,
            True if self._embed_password_flag is not None else None
        ]

    @property
    def optional_parameter_values(self):
        return [
            self._server_address,
            self._port,
            self._connection_username,
            self._connection_password,
            self._embed_password_flag
        ]

    def _validate_inputs(self):
        valid = True
        if self._embed_password_flag:
            if not(self._connection_username and self._connection_password):
                valid = False
        else:
            self._connection_password = ''
        if not valid:
            self._invalid_parameter_exception()

    def base_update_datasource_connection_request(self):
        self._request_body.update({'connection': {}})
        return self._request_body

    def modified_update_datasource_connection_request(self):
        if any(self.optional_parameter_values_exist):
            self._request_body['connection'].update(
                self._get_parameters_dict(self.optional_parameter_keys,
                                          self.optional_parameter_values))
        return self._request_body

    @staticmethod
    def _get_parameters_dict(param_keys, param_values):
        """Override the inherited _get_parameters_dict() method to allow passing boolean values directly"""
        params_dict = {}
        for i, key in enumerate(param_keys):
            if param_values[i] is not None:
                params_dict.update({key: param_values[i]})
        return params_dict

    def get_request(self):
        return self.modified_update_datasource_connection_request()



from tableau_api_lib.api_requests import BaseRequest


class UpdateDatasourceRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests updating datasources.
    :param class ts_connection: the Tableau Server connection object
    :param str new_project_id: (optional) the ID of the project to add the data source to
    :param str new_owner_id: (optional) the ID of the user who will own the datasource
    :param bool is_certified_flag: (optional) True if the datasource is certified, False otherwise
    :param str certification_note: (optional) a note that provides more information on the certification of the
    datasource, if applicable
    """
    def __init__(self,
                 ts_connection,
                 new_project_id=None,
                 new_owner_id=None,
                 is_certified_flag=None,
                 certification_note=None):

        super().__init__(ts_connection)
        self._new_project_id = new_project_id
        self._new_owner_id = new_owner_id
        self._is_certified_flag = is_certified_flag
        self._certification_note = certification_note
        self.base_update_datasource_request()

    @property
    def optional_datasource_param_keys(self):
        return [
            'isCertified',
            'certificationNote'
        ]

    @property
    def optional_project_param_keys(self):
        return ['id']

    @property
    def optional_owner_param_keys(self):
        return ['id']

    @property
    def optional_datasource_param_values(self):
        return [
            'true' if self._is_certified_flag is True else 'false' if self._is_certified_flag is False else None,
            self._certification_note
        ]

    @property
    def optional_project_param_values(self):
        return [self._new_project_id]

    @property
    def optional_owner_param_values(self):
        return [self._new_owner_id]

    def base_update_datasource_request(self):
        self._request_body.update({'datasource': {}})
        return self._request_body

    def modified_update_datasource_request(self):
        if any(self.optional_datasource_param_keys):
            self._request_body['datasource'].update(
                self._get_parameters_dict(self.optional_datasource_param_keys,
                                          self.optional_datasource_param_values))

        if any(self.optional_project_param_keys):
            self._request_body['datasource'].update({'project': {}})
            self._request_body['datasource']['project'].update(
                self._get_parameters_dict(self.optional_project_param_keys,
                                          self.optional_project_param_values))

        if any(self.optional_owner_param_keys):
            self._request_body['datasource'].update({'owner': {}})
            self._request_body['datasource']['owner'].update(
                self._get_parameters_dict(self.optional_owner_param_keys,
                                          self.optional_owner_param_values))

        return self._request_body

    def get_request(self):
        return self.modified_update_datasource_request()



from tableau_api_lib.api_requests import BaseRequest


class AddFlowToScheduleRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests adding flows to schedules.
    :param class ts_connection: the Tableau Server connection object
    :param str flow_id: the flow ID for the flow being added to a schedule
    """
    def __init__(self,
                 ts_connection,
                 flow_id):
        super().__init__(ts_connection)
        self._flow_id = flow_id
        self.base_add_flow_request()

    @property
    def required_flow_param_keys(self):
        return ['id']

    @property
    def required_flow_param_values(self):
        return [self._flow_id]

    def base_add_flow_request(self):
        self._request_body.update({'task': {'flowRun': {}}})
        return self._request_body

    def modified_add_flow_request(self):
        self._request_body['task']['flowRun'].update({'flow': {}})
        self._request_body['task']['flowRun']['flow'].update(
            self._get_parameters_dict(self.required_flow_param_keys,
                                      self.required_flow_param_values))
        return self._request_body

    def get_request(self):
        return self.modified_add_flow_request()



from tableau_api_lib.api_requests import BaseRequest


class AddFlowPermissionsRequest(BaseRequest):
    """
    Builds the request body for Tableau Server REST API requests adding flow permissions.
    :param class ts_connection: the Tableau Server connection object
    :param dict user_capability_dict: the dict defining user capabilities / permissions
    :param dict group_capability_dict: the dict defining group capabilities / permissions
    :param str flow_id: the flow ID for the flow whose permissions are being modified
    :param str user_id: the user ID
    :param str group_id: the group ID
    """
    def __init__(self,
                 ts_connection,
                 user_capability_dict=None,
                 group_capability_dict=None,
                 flow_id=None,
                 user_id=None,
                 group_id=None):

        super().__init__(ts_connection)
        self._flow_id = flow_id
        self._user_id = user_id
        self._group_id = group_id
        self._user_capability_dict = user_capability_dict
        self._group_capability_dict = group_capability_dict
        self._user_capability_names, self._user_capability_modes = None, None
        self._group_capability_names, self._group_capability_modes = None, None
        self._validate_inputs()
        self.base_add_permissions_request()

    @property
    def valid_capability_names(self):
        return [
            'ChangeHierarchy',
            'ChangePermissions',
            'Delete',
            'ExportXml',
            'Read',
            'Write'
        ]

    @property
    def valid_capability_modes(self):
        return [
            'Allow',
            'Deny'
        ]

    def _validate_inputs(self):
        valid = True
        if not (self._user_id or self._group_id):
            valid = False
        if self._user_capability_dict or self._group_capability_dict:
            self._set_capability_variables()
        if not valid:
            self._invalid_parameter_exception()

    def _unpack_capability_dict(self, capability_dict):
        capability_names = []
        capability_modes = []
        for key in capability_dict.keys():
            if key in self.valid_capability_names and capability_dict[key] in self.valid_capability_modes:
                capability_names.append(key)
                capability_modes.append(capability_dict[key])
            else:
                self._invalid_parameter_exception()
        return capability_names, capability_modes

    def _set_capability_variables(self):
        if self._user_capability_dict:
            if any(self._user_capability_dict.values()):
                self._user_capability_names, self._user_capability_modes = self._unpack_capability_dict(
                    self._user_capability_dict)
        if self._group_capability_dict:
            if any(self._group_capability_dict.values()):
                self._group_capability_names, self._group_capability_modes = self._unpack_capability_dict(
                    self._group_capability_dict)

    @staticmethod
    def _get_capability_parameters_list(param_keys, param_values):
        params_list = []
        for i, key in enumerate(param_keys):
            if param_values[i]:
                params_list.append({'name': key,
                                    'mode': param_values[i]})
        return params_list

    @property
    def optional_flow_param_keys(self):
        return ['id']

    @property
    def optional_flow_param_values(self):
        return [self._flow_id]

    def base_add_permissions_request(self):
        self._request_body.update({'permissions': {'granteeCapabilities': []}})
        return self._request_body

    def modified_add_permissions_request(self):
        if any(self.optional_flow_param_values):
            self._request_body['permissions'].update({'flow': {}})
            self._request_body['permissions']['flow'].update(
                self._get_parameters_dict(self.optional_flow_param_keys,
                                          self.optional_flow_param_values))
        if self._user_capability_names:
            capability_dict = {}
            capability_dict.update({'user': {'id': self._user_id}})
            capability_dict.update({'capabilities': {
                'capability': self._get_capability_parameters_list(self._user_capability_names,
                                                                   self._user_capability_modes)
            }})
            self._request_body['permissions']['granteeCapabilities'].append(capability_dict)

        if self._group_capability_names:
            capability_dict = {}
            capability_dict.update({'group': {'id': self._group_id}})
            capability_dict.update({'capabilities': {
                'capability': self._get_capability_parameters_list(self._group_capability_names,
                                                                   self._group_capability_modes)
            }})
            self._request_body['permissions']['granteeCapabilities'].append(capability_dict)

        return self._request_body

    def get_request(self):
        return self.modified_add_permissions_request()
